
import http.client as client
import os
import unittest

from bamboo import App, Endpoint, data_format
from bamboo.api import JsonApiData
from bamboo.base import ContentType, MediaTypes
from bamboo.request import Response
from bamboo.test import ServerForm, TestExecutor
from bamboo.util.time import get_datetime_rfc822


app = App()


class InfoResponse(JsonApiData):
    
    server_name: str
    current_time: str


@app.route("mock", "info")
class MockInfoEndpoint(Endpoint):
    
    NAME_SERVER = "Mocker"
    
    @data_format(input=None, output=InfoResponse)
    def do_GET(self) -> None:
        body = {
            "server_name": self.NAME_SERVER,
            "current_time": get_datetime_rfc822()
        }
        self.send_json(body)


@app.route("mock", "image")
class MockImageEndpoint(Endpoint):
    
    PATH_IMAGE = "elephant.jpg"

    def do_GET(self) -> None:
        with open(self.PATH_IMAGE, "rb") as f:
            image = f.read()
            
        self.send_body(image, ContentType(MediaTypes.jpeg))


class TestResponse(unittest.TestCase):
    
    def setUp(self) -> None:
        self.conn = client.HTTPConnection("localhost", 8000)
        self.url_info = "/" + os.path.join("mock", "info")
        self.url_image = "/" + os.path.join("", "mock", "image")
        self.path_image_ideal = "elephant.jpg"
        
        form = ServerForm("", 8000, app, "test_response.log")
        self.executor = TestExecutor(form).start_serve()
    
    def tearDown(self) -> None:
        self.executor.close()

    def test_body(self):
        with open(self.path_image_ideal, "rb") as f:
            image_ideal = f.read()
        
        self.conn.request("GET", self.url_image)
        _res = self.conn.getresponse()
        with Response(self.conn, _res) as res:
            self.assertEqual(image_ideal, res.body)
        
    def test_attach_datacls(self):
        self.conn.request("GET", self.url_info)
        _res = self.conn.getresponse()
        with Response(self.conn, _res, datacls=InfoResponse) as res:
            data = res.attach()
        
        self.assertTrue(isinstance(data, InfoResponse))
        self.assertEqual(data.server_name, "Mocker")
        print(f"Response of 'current_time' : {data.current_time}")


if __name__ == "__main__":
    unittest.main()




from abc import ABCMeta, abstractmethod
from typing import Optional, Tuple, Union


class FlexibleLocation(metaclass=ABCMeta):
    """Base class of flexible location.
    
    Location is concept representing each pieces of path of URI, i.e. 
    locations can configures a path of URI by joining them with 
    separator `/`. 
    
    Location has type of `str` or subclasses of this class. If a location 
    is `str`, then the location is called a static location. Otherwise, 
    the location is called a flexible location. Flexible locations means 
    location with some kind of logical rules of strings. The logic of rules 
    can be implemented on `is_valid` method, returning if specified string 
    is valid in the rules. 
    """
    
    @abstractmethod
    def is_valid(self, loc: str) -> bool:
        pass


Location_t = Union[str, FlexibleLocation]
Uri_t = Tuple[Location_t]


def is_flexible_uri(uri: Uri_t) -> bool:
    """Judge if specified `uri` has one or more flexible location.

    Parameters
    ----------
    uri : Uri_t
        URI pattern to be judged

    Returns
    -------
    bool
        If specified `uri` has one or more flexible location.
    """
    for loc in uri:
        if isinstance(loc, FlexibleLocation):
            return True
    return False


def is_duplicated_uri(uri_1: Uri_t, uri_2: Uri_t) -> bool:
    """Judge if a couple of specified URI patterns has same pattern.

    Parameters
    ----------
    uri_1 : Uri_t
        URI pattern to be judged
    uri_2 : Uri_t
        URI pattern to be judged

    Returns
    -------
    bool
        If two URIs has same pattern or not
    """
    if len(uri_1) == len(uri_2):
        for loc_1, loc_2 in zip(uri_1, uri_2):
            if isinstance(loc_1, FlexibleLocation):
                continue
            if isinstance(loc_2, FlexibleLocation):
                continue
            
            if loc_1 != loc_2:
                break
        else:
            return True
    return False


class AsciiDigitLocation(FlexibleLocation):
    """Flexible location representing location of number in ASCII code.
    """
    
    def __init__(self, digits: int) -> None:
        """
        Parameters
        ----------
        digits : int
            Number of digits which the location accepts

        Raises
        ------
        ValueError
            Raised if `digits` is 0 or less
        """
        if digits < 1:
            raise ValueError("'digits' must be bigger than 0.")
        
        self._digits = digits
        
    def is_valid(self, loc: str) -> bool:
        """Judge if specified `loc` is valid location or not.

        Parameters
        ----------
        loc : str
            Location to be judged

        Returns
        -------
        bool
            If specified location is valid or not
        """
        return loc.isascii() and loc.isdigit() and len(loc) == self._digits
        
        
class AnyStringLocation(FlexibleLocation):
    """Flexible location representing string with no rules.
    """
    
    def __init__(self, max: Optional[int] = None) -> None:
        """
        Parameters
        ----------
        max : Optional[int], optional
            Max length of string of location, by default `None`

        Raises
        ------
        ValueError
            Raised if `max` is 0 or less
            
        Notes
        -----
        If the argument `max` is `None`, then any length of string 
        will be accepted.
        """
        if max and max < 1:
            raise ValueError("'max' must be bigger than 0.")

        self._max = max
        
    def is_valid(self, loc: str) -> bool:
        """Judge if specified `loc` is valid location or not.

        Parameters
        ----------
        loc : str
            Location to be judged

        Returns
        -------
        bool
            If specified location is valid or not
        """
        if self._max is None:
            return True
        
        len_loc = len(loc)
        if len_loc > self._max or len_loc == 0:
            return False
        return True





from __future__ import annotations
from dataclasses import dataclass
from multiprocessing import Process
import signal
import sys
import time
from typing import Any, Callable, List, Tuple
from wsgiref.simple_server import make_server

from bamboo.app import App
from bamboo.util.string import ColorCode, insert_colorcode


@dataclass
class ServerForm:
    """Dataclass to register a server form.
    
    This dataclass would be used to register as a form 
    into test classes such as TestExecutor.
    
    Parameters
    ----------
    host : str
        Hostname of the server
    port : int
        Port of the server
    app : App
        App object with implemented Endpoints
    path_log : str
        Path log of the server will be written
    """
    host: str
    port: int
    app: App
    path_log: str
    
    
def serve_at(form: ServerForm) -> None:
    """Subroutine for server application called at a child process.

    Parameters
    ----------
    form : ServerForm
        Dataclass describing information of the server application
    """
    server = make_server(form.host, form.port, form.app)
    f_log = open(form.path_log, "wt")
    sys.stdout = f_log
    sys.stderr = f_log
    
    def server_close(signalnum, frame):
        server.server_close()
        print()
        f_log.close()
        sys.exit()
    
    signal.signal(signal.SIGTERM, server_close)
    server.serve_forever()
    
    
class TestExecutor:
    """Utility class that can execute server applications at child processes.
    
    This class can be used in test scripts for holding several server forms 
    and running the application based on the forms at the child processes. 
    Also this object has feature of context manager and developers may use 
    it to kill the processes safely.
    """
    
    def __init__(self, *forms: ServerForm) -> None:
        self._forms: List[ServerForm] = []
        self._children: List[Process] = []

        self.add_forms(*forms)
        
    def add_forms(self, *forms: ServerForm) -> None:
        """Add forms with information of server applications.
        
        Parameters
        ----------
        *foms : ServerForm
            Dataclass describing information of the server application
        """
        for form in forms:
            self._forms.append(form)
            
    def start_serve(self, waiting: float = 0.05) -> TestExecutor:
        """Run registered server applications at child processes.
        
        This object has feature of context manager and the method 
        returns the object itself. So developer can use the with 
        sentence and in it, can define logic of clients.

        Parameters
        ----------
        waiting : float, optional
            Waiting time after running the processes, by default 0.05

        Returns
        -------
        TestExecutor
            This object itself
            
        Examples
        --------
        ```python
        >>> holder = TestExecutor(form)
        >>> with holder.start_serve():
        ...    res = http_get("http://localhost:8000/image")
        >>> print(res.body)
        ```
        """
        for form in self._forms:
            child = Process(target=serve_at, args=(form,))
            child.start()
            self._children.append(child)
        time.sleep(waiting)
        
        return self
        
    def close(self, pop: bool = True) -> None:
        """Kill the all child processes derived from registered forms.

        Parameters
        ----------
        pop : bool, optional
            If removing the registered forms, by default True
        """
        for child in self._children:
            child.terminate()
            child.join()
            child.close()
            
        if pop:
            self._children.clear()
            
    def exec(self, func: Callable[[Tuple[Any, ...]], None], 
             args: Tuple[Any, ...] = (), waiting: float = 0.05) -> None:
        """Executes a simple client-server test.

        Parameters
        ----------
        func : Callable[[Tuple[Any, ...]], None]
            Function executed after all the server applications start
        args : Tuple[Any, ...], optional
            Arguments of the func, by default ()
        waiting : float, optional
            Waiting time after running the applications, by default 0.05
        """
        with self.start_serve(waiting=waiting):
            func(*args)
           
    @classmethod 
    def debug(cls, app: App, path_log: str, host: str = "localhost", 
              port: int = 8000, waiting: float = 0.05) -> None:
        """Executes a server application for debug.
        
        This method is a kind of shorcut for launching a server application 
        and can be used to debug the application. If you want to deploy an
        application made with Bamboo, consider to use another WSGI server 
        application for production.

        Parameters
        ----------
        app : App
            App object with implemented Endpoints
        path_log : str
            Path log of the server will be written
        host : str, optional
            Hostname of the server, by default "localhost"
        port : int, optional
            Port of the server, by default 8000
        waiting : float, optional
            Waiting time after running the applications, by default 0.05
        """
        form = ServerForm(host, port, app, path_log)
        executor = cls(form)
        
        try:
            print(f"Hosting on {host}:{port} ...")
            print(insert_colorcode(
                "WARNING: This is debug mode. "
                "Do not use it in your production deployment.",
                ColorCode.RED
            ))
            
            executor.start_serve(waiting=waiting)
            while True:
                time.sleep(60 * 60)
        except KeyboardInterrupt:
            executor.close()
            print("\nHosting terminated. The log was output into " +
                  insert_colorcode(f"{path_log}", ColorCode.YELLOW) + ".")
            
    def __enter__(self) -> None:
        pass
    
    def __exit__(self, type, value, traceback) -> None:
        self.close()




from __future__ import annotations
from dataclasses import dataclass
import inspect
import json
from typing import (
    Any, Callable, Dict, List, Optional, Tuple, Type,
)
from urllib.parse import parse_qs

from bamboo.api import ApiData, ValidationFailedError
from bamboo.base import (
    HTTPMethods, HTTPStatus, MediaTypes, ContentType,
    AuthSchemes, DEFAULT_CONTENT_TYPE_PLAIN,
)
from bamboo.error import (
    DEFAULT_BASIC_AUTH_HEADER_NOT_FOUND_ERROR, 
    DEFAULT_BEARER_AUTH_HEADER_NOT_FOUND_ERROR, 
    DEFAULT_HEADER_NOT_FOUND_ERROR, DEFAULT_NOT_APPLICABLE_IP_ERROR,
    DEFUALT_INCORRECT_DATA_FORMAT_ERROR, ErrInfoBase,
)
from bamboo.util.convert import decode2binary
from bamboo.util.deco import cached_property
from bamboo.util.ip import is_valid_ipv4


class StatusCodeAlreadySetError(Exception):
    """Raised if response status code has already been set."""
    pass


class Endpoint:
    """Base class of Endpoint to define logic to requests.
    
    Attributes
    ----------
    response_methods : Tuple[str]
        Callbacks defined the class
    """
    
    response_methods: Tuple[str]
    
    @classmethod
    def _get_response_method_names(cls) -> Tuple[str]:
        """Retrieve names of response methods defined on the class.

        Returns
        -------
        Tuple[str]
            Names of response methods.
        """
        result = []
        for name, _ in inspect.getmembers(cls):
            if len(name) < 3:
                continue
            if name[:3] == "do_" and name[3:] in HTTPMethods:
                result.append(name[3:])
        return tuple(result)
    
    def __init_subclass__(cls) -> None:
        cls.response_methods = cls._get_response_method_names()
        
    def __init__(self, 
                 environ: Dict[str, Any], 
                 flexible_locs: Tuple[str, ...],
                 *parcel
                 ) -> None:
        self._environ = environ
        self._flexible_locs = flexible_locs
        self._req_body = None
        self._status: Optional[HTTPStatus] = None
        self._headers: List[Tuple[str, str]] = []
        self._res_body: bytes = b""

        self.setup(*parcel)
        
    @classmethod
    def _get_response_method(cls, method: str) -> Optional[Callback_t]:
        """Retrieve response methods with given HTTP method.

        Parameters
        ----------
        method : str
            HTTP method

        Returns
        -------
        Optional[Callable_t]
            Callback with given name
        """
        mname = "do_" + method
        if hasattr(cls, mname):
            return getattr(cls, mname)
        return None
        
    def _recv_body_secure(self) -> bytes:
        """Receive request body securely.

        Returns
        -------
        bytes
            Request body
        """
        # TODO
        #   Take measures against DoS attack.
        body = self._environ.get("wsgi.input").read(self.content_length)
        return body
        
    def setup(self, *parcel) -> None:
        pass
    
    @property
    def flexible_locs(self) -> Tuple[str, ...]:
        return self._flexible_locs
    
    @property
    def client_ip(self) -> str:
        return self._environ.get("REMOTE_ADDR")
    
    @property
    def requested_addr(self) -> Tuple[str, int]:
        """Retrieve requested a pair of host name and port.

        Returns
        -------
        Tuple[str, int]
            A pair of host name and port
        """
        host = self._environ.get("HTTP_HOST")
        if host:
            host = host.split(":")[0]
        else:
            host = self._environ.get("SERVER_NAME")
        
        port = self._environ.get("SERVER_PORT")
        if len(port):
            port = int(port)
        
        return (host, port)
        
    def get_header(self, name: str) -> Optional[str]:
        """Try to retrieve a HTTP header.

        Parameters
        ----------
        name : str
            Header field to retrieve

        Returns
        -------
        Optional[str]
            Value of the field if exists, else None
        """
        name = "HTTP_" + name.replace("-", "_").upper()
        return self._environ.get(name)
    
    @cached_property
    def content_type(self) -> ContentType:
        raw = self._environ.get("CONTENT_TYPE")
        return ContentType.parse(raw)
    
    @property
    def content_length(self) -> int:
        length = self._environ.get("CONTENT_LENGTH")
        if length:
            return int(length)
        return 0
    
    @property
    def path(self) -> str:
        return self._environ.get("PATH_INFO")

    @property
    def request_method(self) -> str:
        return self._environ.get("REQUEST_METHOD").upper()
    
    @property
    def query(self) -> Dict[str, str]:
        return parse_qs(self._environ.get("QUERY_STRING"))
    
    @property
    def body(self) -> bytes:
        if self._req_body is None:
            self._req_body = self._recv_body_secure()
        return self._req_body
    
    def add_header(self, name: str, value: str, 
                   **params: Optional[str]) -> None:
        """Add response header with MIME parameters.

        Parameters
        ----------
        name : str
            Field name of the header
        value : str
            Value of the field
        **params : Optional[str]
            MIME parameters added to the field
        """
        params = [f'; {header}={val}' for header, val in params.items()]
        params = "".join(params)
        self._headers.append((name, value + params))
    
    def add_headers(self, **headers: str) -> None:
        """Add response headers at once.

        Parameters
        ----------
        **headers : str
            Header's info whose header is the field name.
            
        Notes
        -----
        This method would be used shortcut to register multiple 
        headers. If it requires adding MIME parameters, developers
        can use the 'add_header' method.
        """
        for name, val in headers.items():
            self.add_header(name, val)
    
    def _check_status_already_set(self) -> None:
        """Check if response status code already exists.

        Raises
        ------
        StatusCodeAlreadySetError
            Raised if response status code has already been set.
        """
        if self._res_body:
            raise StatusCodeAlreadySetError(
                "Response status code has already been set.")
        
    def send_only_status(self, status: HTTPStatus = HTTPStatus.OK) -> None:
        """Set specified status code to one of response.
        
        This method can be used if a callback doesn't need to send response 
        body. 

        Parameters
        ----------
        status : HTTPStatus, optional
            HTTP status of the response, by default `HTTPStatus.OK`
        """
        self._check_status_already_set()
        
        self._status = status
    
    def send_body(self, body: bytes, 
                  content_type: ContentType = DEFAULT_CONTENT_TYPE_PLAIN,
                  status: HTTPStatus = HTTPStatus.OK) -> None:
        """Set given binary to the response body.

        Parameters
        ----------
        body : bytes, optional
            Binary to be set to the response body, by default `b""`
        content_type : ContentType, optional
            `Content-Type` header to be sent, 
            by default `DEFAULT_CONTENT_TYPE_PLAIN`
        status : HTTPStatus, optional
            HTTP status of the response, by default `HTTPStatus.OK`
            
        Notes
        -----
        If the parameter `content_type` is specified, then the `Content-Type` 
        header is to be added.
        
        `DEFAULT_CONTENT_TYPE_PLAIN` has its MIME type of `text/plain`, and the 
        other attributes are `None`. If another value of `Content-Type` is 
        needed, then you should generate new `ContentType` instance with 
        attributes you want.
            
        Raisess
        ------
        StatusCodeAlreadySetError
            Raised if response status code has already been set.
        """
        self._check_status_already_set()

        self._status = status
        self._res_body = body
        
        # Content-Type's parameters
        params = {}
        if content_type.charset:
            params["charset"] = content_type.charset
        if content_type.boundary:
            params["boundary"] = content_type.boundary    
    
        self.add_header("Content-Type", content_type.media_type, **params)
    
    def send_json(self, body: Dict[str, Any], 
                  status: HTTPStatus = HTTPStatus.OK,
                  encoding: str = "UTF-8") -> None:
        """Set given json data to the response body.

        Parameters
        ----------
        body : Dict[str, Any]
            Json data to be set to the response body
        status : HTTPStatus, optional
            HTTP status of the response, by default HTTPStatus.OK
        encoding : str, optional
            Encoding of the Json data, by default "UTF-8"
            
        Raises
        ------
        StatusCodeAlreadySetError
            Raised if response status code has already been set.
        """
        self._check_status_already_set()
        
        body = json.dumps(body).encode(encoding=encoding)
        content_type = ContentType(media_type=MediaTypes.json, charset=encoding)
        self.send_body(body, content_type=content_type, status=status)
    
    def send_err(self, err: ErrInfoBase) -> None:
        """Set error to the response body.

        Parameters
        ----------
        err : ErrInfoBase
            Error information to be sent
            
        Raises
        ------
        StatusCodeAlreadySetError
            Raised if response status code has already been set.
        """
        self._check_status_already_set()

        self._status = err.http_status
        self._res_body = err.get_body()
        
        for name, val in err.get_headers():
            self.add_header(name, val)
        
        if len(self._res_body):
            self.add_header("Content-Type", err._content_type_,
                            **err._content_type_args_)


# Signature of the callback of response method on sub classes of 
# the Endpoint class.
Callback_t = Callable[[Endpoint, Tuple[Any, ...]], None]
CallbackDecorator_t = Callable[[Callback_t], Callback_t]

def _get_bamboo_attr(attr: str) -> str:
    return f"__bamboo_{attr}__"


# Error Information ----------------------------------------------------------

ATTR_ERRORS = _get_bamboo_attr("errors")


def get_errors_info(callback: Callback_t) -> List[Type[ErrInfoBase]]:
    """Retrieve errors may occur at specified `callback`.

    Parameters
    ----------
    callback : Callback_t
        Response method implemented on `Endpoint`

    Returns
    -------
    List[Type[ErrInfoBase]]
        `list` of error classes may occur
    """
    if hasattr(callback, ATTR_ERRORS):
        return getattr(callback, ATTR_ERRORS)
    return []


def may_occur(*errors: Type[ErrInfoBase]) -> CallbackDecorator_t:
    """Register error classes to callback on `Endpoint`.
    
    Parameters
    ----------
    *errors : Type[ErrInfoBase]
        Error classes which may occuur

    Returns
    -------
    CallbackDecorator_t
        Decorator to register error classes to callback
        
    Examples
    --------
    ```
    class MockErrInfo(ErrInfo):
        http_status = HTTPStatus.INTERNAL_SERVER_ERROR
        
        @classmethod
        def get_body(cls) -> bytes:
            return b"Intrernal server error occured"
            
    class MockEndpoint(Endpoint):
        
        @may_occur(MockErrInfo)
        def do_GET(self) -> None:
            # Do something...
            
            # It is possible to send error response.
            if is_some_flag():
                self.send_err(MockErrInfo)

            self.send_body(status=HTTPStatus.OK)
    ```
    """
    def attach_err_info(callback: Callback_t) -> Callback_t:
        if not hasattr(callback, ATTR_ERRORS):
            setattr(callback, ATTR_ERRORS, set())
        
        registered = getattr(callback, ATTR_ERRORS)
        for err in errors:
            registered.add(err)
            
        return callback
    return attach_err_info

# ----------------------------------------------------------------------------

# Data Format   --------------------------------------------------------------

ATTR_DATA_FORMAT = _get_bamboo_attr("data_format")


@dataclass
class DataFormatInfo:
    """`dataclass` with information of data format at callbacks on `Endpoint`.
    
    Attributes
    ----------
    input : Optional[Type[ApiData]]
        Input data format, by default `None`
    output : Optional[Type[ApiData]]
        Output data format, by default `None`
    is_validate : bool
        If input data is to be validate, by default `True`
    err_validate : ErrInfoBase
        Error information sent when validation failes, 
        by default `DEFUALT_INCORRECT_DATA_FORMAT_ERROR`
    """
    input: Optional[Type[ApiData]] = None
    output: Optional[Type[ApiData]] = None
    is_validate: bool = True
    err_validate: ErrInfoBase = DEFUALT_INCORRECT_DATA_FORMAT_ERROR
    
    
def get_data_format_info(callback: Callback_t) -> Optional[DataFormatInfo]:
    """Retrieve information of data format at `callback` on `Endpoint`.

    Parameters
    ----------
    callback : Callback_t
        Response method implemented on `Endpoint`

    Returns
    -------
    Optional[DataFormatInfo]
        Information of data format at the `callback`. If the `callback` is 
        not decorated by `data_format` decorator, then returns `None`.
    """
    if hasattr(callback, ATTR_DATA_FORMAT):
        info = getattr(callback, ATTR_DATA_FORMAT)
        return DataFormatInfo(**info)
    return None
    

def data_format(input: Optional[Type[ApiData]] = None, 
                output: Optional[Type[ApiData]] = None,
                is_validate: bool = True,
                err_validate: ErrInfoBase = 
                DEFUALT_INCORRECT_DATA_FORMAT_ERROR) -> CallbackDecorator_t:
    """Set data format of input/output data as API to `callback` on 
    `Endpoint`.
    
    This decorator can be used to add attributes of data format information 
    to a `callback`, and execute validation if input raw data has expected 
    format defined on `input` argument. 
    
    To represent no data inputs/outputs, specify `input`/`output` arguments 
    as `None`s. If `input` is `None`, then any data received from client will 
    not be read. If `is_validate` is `False`, then validation will not be 
    executed.
    
    To retrieve data format information, use `get_data_format_info` function.

    Parameters
    ----------
    input : Optional[Type[ApiData]]
        Input data format, by default `None`
    output : Optional[Type[ApiData]]
        Output data format, by default `None`
    is_validate : bool, optional
        If input data is to be validated, by default `True`
    err_validate : ErrInfoBase
        Error information sent when validation failes, 
        by default `DEFUALT_INCORRECT_DATA_FORMAT_ERROR`

    Returns
    -------
    CallbackDecorator_t
        Decorator to add attributes of data format information to callback
        
    Examples
    --------
    ```
    class UserData(JsonApiData):
        name: str
        email: str
        age: int
    
    class MockEndpoint(Endpoint):
    
        @data_format(input=UserData, output=None)
        def do_GET(self, rec_body: UserData) -> None:
            user_name = rec_body.name
            # Do something...
    ```
    """
    _format = {"input": input, "output": output, 
               "err_validate": err_validate, "is_validate": is_validate}
    
    if is_validate and input:
        def input_decoded(
            callback: Callable[[Endpoint, ApiData, Tuple[Any, ...]], None]
            ) -> Callback_t:
            
            @may_occur(err_validate.__class__)
            def _callback(self: Endpoint, *args) -> None:
                body = self.body
                try:
                    data = input(body, self.content_type)
                except ValidationFailedError:
                    self.send_err(err_validate)
                    return
                
                callback(self, data, *args)
                
            _callback.__dict__ = callback.__dict__
            setattr(callback, ATTR_DATA_FORMAT, _format)
                
            return _callback
        return input_decoded
    else:
        def input_decoded(
            callback: Callable[[Endpoint, Tuple[Any, ...]], None]
            ) -> Callback_t:
            setattr(callback, ATTR_DATA_FORMAT, _format)
            
            # NOTE
            #   If input is None, then any response body would be received.
            if input is None:
                def _callback(self: Endpoint, *args) -> None:
                    self._req_body = b""
                    callback(self, *args)
                _callback.__dict__ = callback.__dict__
                return _callback
            
            return callback
        return input_decoded
    
# ----------------------------------------------------------------------------

# Required Headers  ----------------------------------------------------------

ATTR_HEADERS_REQUIRED = _get_bamboo_attr("headers_required")


@dataclass
class RequiredHeaderInfo:
    """`dataclass` with information of header which should be included in 
    response headers.
    
    Attributes
    ----------
    header : str
        Name of header
    err : ErrInfoBase
        Error information sent when the header is not included,
        by default `DEFAULT_HEADER_NOT_FOUND_ERROR`
    """
    header: str
    err: ErrInfoBase = DEFAULT_HEADER_NOT_FOUND_ERROR
    
    
def get_required_header_info(callback: Callback_t
                             ) -> List[RequiredHeaderInfo]:
    """Retrieve information of headers which should be included in 
    response headers.

    Parameters
    ----------
    callback : Callback_t
        Response method implemented on `Endpoint`

    Returns
    -------
    List[RequiredHeaderInfo]
        `list` of information of required headers
    """
    if hasattr(callback, ATTR_HEADERS_REQUIRED):
        arr_info = getattr(callback, ATTR_HEADERS_REQUIRED)
        return [RequiredHeaderInfo(*info) for info in arr_info]
    return []


def has_header_of(header: str, 
                  err: ErrInfoBase = DEFAULT_HEADER_NOT_FOUND_ERROR
                  ) -> CallbackDecorator_t:
    """Set callback up to receive given header from clients.

    If request headers don't include specified `header`, then response 
    headers and body will be made based on `err` and sent.

    Parameters
    ----------
    header : str
        Name of header
    err : ErrInfoBase, optional
        Error information sent when specified `header` is not found of request 
        headers, by default `DEFAULT_HEADER_NOT_FOUND_ERROR`

    Returns
    -------
    CallbackDecorator_t
        Decorator to make callback to be set up to receive the header
        
    Examples
    --------
    ```
    class BasicAuthHeaderNotFoundErrInfo(ErrInfoBase):
        http_status = HTTPStatus.UNAUTHORIZED
        
        @classmethod
        def get_headers(cls) -> List[Tuple[str, str]]:
            return [("WWW-Authenticate", 'Basic realm="SECRET AREA"')]
            
    class MockEndpoint(Endpoint):
    
        @has_header_of("Authorization", BasicAuthHeaderNotFoundErrInfo)
        def do_GET(self) -> None:
            # It is guaranteed that request headers include the 
            # `Authorization` header at this point.
            header_auth = self.get_header("Authorization")
            
            # Do something...
    ```
    """
    def header_checker(callback: Callback_t) -> Callback_t:
        
        @may_occur(err.__class__)
        def _callback(self: Endpoint, *args) -> None:
            val = self.get_header(header)
            if val is None:
                self.send_err(err)
                return
            callback(self, *args)
        
        _callback.__dict__ = callback.__dict__
        if not hasattr(_callback, ATTR_HEADERS_REQUIRED):
            setattr(_callback, ATTR_HEADERS_REQUIRED, set())
        
        headers_required = getattr(_callback, ATTR_HEADERS_REQUIRED)
        headers_required.add((header, err))
        return _callback
    return header_checker

# ----------------------------------------------------------------------------

# IP restriction    ----------------------------------------------------------

ATTR_CLIENT_RESTRICTED = _get_bamboo_attr("client_restricted")


def get_restricted_ip_info(callback: Callback_t) -> List[str]:
    """Retrieve IP addresses restricted at specified `callback`.

    Parameters
    ----------
    callback : Callback_t
        Response method implemented on `Endpoint`

    Returns
    -------
    List[str]
        `list` of restricted IP addresses
    """
    if hasattr(callback, ATTR_CLIENT_RESTRICTED):
        return getattr(callback, ATTR_CLIENT_RESTRICTED)
    return []


def restricts_client(*client_ips: str, 
                     err: ErrInfoBase = DEFAULT_NOT_APPLICABLE_IP_ERROR
                     ) -> CallbackDecorator_t:
    """Restrict IP addresses at callback.

    Parameters
    ----------
    *client_ips : str
        IP addresses to be allowed to request
    err : ErrInfoBase, optional
        Error information sent when request from IP not included specified IPs
         comes, by default `DEFAULT_NOT_APPLICABLE_IP_ERROR`

    Returns
    -------
    CallbackDecorator_t
        Decorator to make callback to be set up to restrict IP addresses
        
    Raises
    ------
    ValueError
        Raised if invalid IP address is detected
        
    Examples
    --------
    ```
    class MockEndpoint(Endpoint):
        
        # Restrict to allow only localhost to request 
        # to this callback
        @restricts_client("localhost")
        def do_GET(self) -> None:
            # Only localhost can access to the callback.
        
            # Do something...
    ```
    """
    for ip in client_ips:
        if not is_valid_ipv4(ip):
            raise ValueError(f"{ip} is invalid IP address.")
    
    allowed_ips = set(client_ips)
    # Convert localhost to the address
    if "localhost" in allowed_ips:
        allowed_ips.remove("localhost")
        allowed_ips.add("127.0.0.1")
    
    def register_restrictions(callback: Callback_t) -> Callback_t:
            
        @may_occur(err.__class__)
        def _callback(self: Endpoint, *args) -> None:
            if self.client_ip not in allowed_ips:
                self.send_err(err)
                return
            callback(self, *args)
            
        _callback.__dict__ = callback.__dict__
        if not hasattr(callback, ATTR_CLIENT_RESTRICTED):
            setattr(callback, ATTR_CLIENT_RESTRICTED, set())
            
        restrictions = getattr(callback, ATTR_CLIENT_RESTRICTED)
        restrictions.update(*allowed_ips)
            
        return _callback
    return register_restrictions

# ----------------------------------------------------------------------------

_HEADER_AUTHORIZATION = "Authorization"
ATTR_AUTH_SCHEME = _get_bamboo_attr("auth_scheme")


class MultipleAuthSchemeError(Exception):
    """Raised if several authentication schemes of the 'Authorization' 
    header are detected.
    """
    pass


def get_auth_scheme(callback: Callback_t) -> Optional[str]:
    return getattr(callback, ATTR_AUTH_SCHEME, None)
    
    
def _set_auth_scheme(callback: Callback_t, scheme: str) -> None:
    if hasattr(callback, ATTR_AUTH_SCHEME):
        _scheme_registered = getattr(callback, ATTR_AUTH_SCHEME)
        raise MultipleAuthSchemeError(
            "Authentication scheme has been specified as "
            f"'{_scheme_registered}'. Do not specify multiple schemes."
        )
    setattr(callback, ATTR_AUTH_SCHEME, scheme)
    
    
def _validate_auth_header(value: str, scheme: str) -> Optional[str]:
    value = value.split(" ")
    if len(value) != 2:
        return None
    
    _scheme, credentials = value
    if _scheme != scheme:
        return None
    
    return credentials
    

def basic_auth(err: ErrInfoBase = DEFAULT_BASIC_AUTH_HEADER_NOT_FOUND_ERROR
               ) -> CallbackDecorator_t:
    """Set callback up to require `Authorization` header in Basic 
    authentication.

    Parameters
    ----------
    err : ErrInfoBase, optional
        Error sent when `Authorization` header is not found, received 
        scheme doesn't match, or extracting user ID and password from 
        credentials failes, 
        by default DEFAULT_BASIC_AUTH_HEADER_NOT_FOUND_ERROR

    Returns
    -------
    CallbackDecorator_t
        Decorator to make callback to be set  up to require 
        the `Authorization` header
        
    Examples
    --------
    ```
    class MockEndpoint(Endpoint):
    
        @basic_auth()
        def do_GET(self, user_id: str, password: str) -> None:
            # It is guaranteed that request headers include the 
            # `Authorization` header at this point, and user_id and 
            # password are the ones extracted from the header.
            
            # Authenticate with any function working on your system.
            authenticate(user_id, password)
            
            # Do something...
    ```
    """
    def auth_header_checker(callback: Callback_t) -> Callback_t:
        
        @may_occur(err.__class__)
        @has_header_of(_HEADER_AUTHORIZATION, err)
        def _callback(self: Endpoint, *args) -> None:
            val = self.get_header(_HEADER_AUTHORIZATION)
            credentials = _validate_auth_header(val, AuthSchemes.basic)
            if credentials is None:
                self.send_err(err)
                return
            
            credentials = decode2binary(credentials).decode().split(":")
            if len(credentials) != 2:
                self.send_err(err)
                return
            
            user_id, pw = credentials
            callback(self, user_id, pw)
            
        _callback.__dict__ = callback.__dict__
        _set_auth_scheme(callback, AuthSchemes.basic)
        return _callback
    return auth_header_checker


def bearer_auth(err: ErrInfoBase = DEFAULT_BEARER_AUTH_HEADER_NOT_FOUND_ERROR
                ) -> CallbackDecorator_t:
    """Set callback up to require `Authorization` header in token 
    authentication for OAuth 2.0.

    Parameters
    ----------
    err : ErrInfoBase, optional
        Error sent when `Authorization` header is not found, or when received 
        scheme doesn't match, 
        by default DEFAULT_BEARER_AUTH_HEADER_NOT_FOUND_ERROR

    Returns
    -------
    CallbackDecorator_t
        Decorator to make callback to be set  up to require 
        the `Authorization` header
        
    Examples
    --------
    ```
    class MockEndpoint(Endpoint):
    
        @bearer_auth()
        def do_GET(self, token: str) -> None:
            # It is guaranteed that request headers include the 
            # `Authorization` header at this point, and token is
            # the one extracted from the header.
            
            # Authenticate with any function working on your system.
            authenticate(token)
            
            # Do something...
    ```
    """
    def auth_header_checker(callback: Callback_t) -> Callback_t:
        
        @may_occur(err.__class__)
        @has_header_of(_HEADER_AUTHORIZATION, err)
        def _callback(self: Endpoint, *args) -> None:
            val = self.get_header(_HEADER_AUTHORIZATION)
            token = _validate_auth_header(val, AuthSchemes.bearer)
            if token is None:
                self.send_err(err)
                return
            
            callback(self, token)
            
        _callback.__dict__ = callback.__dict__
        _set_auth_scheme(callback, AuthSchemes.bearer)
        return _callback
    return auth_header_checker




from typing import List
import unittest

from bamboo.api import (
    JsonApiDataBuilder, JsonApiData, InvalidAnnotationError
)


class TestInnerApi(JsonApiData):
    
    name: str
    age: int
    
    
class TestOuterApi(JsonApiData):
    
    accounts: List[TestInnerApi]
    

class TestJsonable(unittest.TestCase):
    
    def test_int(self):
        JsonApiDataBuilder.check_annotations(int)
    
    def test_float(self):
        JsonApiDataBuilder.check_annotations(float)
    
    def test_str(self):
        JsonApiDataBuilder.check_annotations(str)
    
    def test_bool(self):
        JsonApiDataBuilder.check_annotations(bool)
    
    def test_NoneType(self):
        JsonApiDataBuilder.check_annotations(type(None))
    
    def test_list(self):
        JsonApiDataBuilder.check_annotations(List[int])

        with self.assertRaises(InvalidAnnotationError) as err:
            JsonApiDataBuilder.check_annotations(List)
        self.assertIsInstance(err.exception, InvalidAnnotationError)
        
        with self.assertRaises(InvalidAnnotationError) as err:
            JsonApiDataBuilder.check_annotations(List[dict])
        self.assertIsInstance(err.exception, InvalidAnnotationError)
    
    def test_jsonapi(self):
        JsonApiDataBuilder.has_valid_annotations(TestOuterApi)


if __name__ == "__main__":
    unittest.main()




from __future__ import annotations
from http.client import HTTPSConnection
from ssl import SSLContext
from typing import Any, Dict, List, Optional, Type

from bamboo.api import BinaryApiData
from bamboo.base import HTTPMethods
from bamboo.request import ResponseData_t, Schemes
from bamboo.request.request_form import get_http_request_form
from bamboo.request.response import Response


def request(uri: str,
            method: str,
            headers: Dict[str, str] = {},
            body: Optional[bytes] = None,
            json: Optional[Dict[str, Any]] = None,
            query: Dict[str, List[str]] = {},
            timeout: Optional[float] = None,
            blocksize: int = 8192,
            datacls: Type[ResponseData_t] = BinaryApiData,
            key_file: Optional[str] = None,
            cert_file: Optional[str] = None,
            context: Optional[SSLContext] = None,
            ) -> Response[ResponseData_t]:    
    form = get_http_request_form(Schemes.HTTPS, uri, method, headers=headers,
                                 body=body, json=json, query=query)
    conn = HTTPSConnection(form.host, form.port, key_file=key_file,
                           cert_file=cert_file, context=context,
                           timeout=timeout, blocksize=blocksize)
    conn.request(form.method, form.uri, body=form.body, headers=form.headers)
    _res = conn.getresponse()
    return Response(conn, _res, datacls=datacls)


def get(uri: str, 
        headers: Dict[str, str] = {}, 
        body: Optional[bytes] = None,
        json: Optional[Dict[str, Any]] = None,
        query: Dict[str, List[str]] = {},
        timeout: Optional[float] = None, 
        blocksize: int = 8192,
        datacls: Type[ResponseData_t] = BinaryApiData
        ) -> Response[ResponseData_t]:
    return request(uri, HTTPMethods.GET, headers=headers, body=body,
                   json=json, query=query, timeout=timeout,
                   blocksize=blocksize, datacls=datacls)


def post(uri: str, 
        headers: Dict[str, str] = {},
        body: Optional[bytes] = None,
        json: Optional[Dict[str, Any]] = None,
        query: Dict[str, List[str]] = {},
        timeout: Optional[float] = None, 
        blocksize: int = 8192,
        datacls: Type[ResponseData_t] = BinaryApiData
        ) -> Response[ResponseData_t]:
    return request(uri, HTTPMethods.POST, headers=headers, body=body,
                   json=json, query=query, timeout=timeout,
                   blocksize=blocksize, datacls=datacls)


def put(uri: str, 
        headers: Dict[str, str] = {},
        body: Optional[bytes] = None,
        json: Optional[Dict[str, Any]] = None,
        query: Dict[str, List[str]] = {},
        timeout: Optional[float] = None, 
        blocksize: int = 8192,
        datacls: Type[ResponseData_t] = BinaryApiData
        ) -> Response[ResponseData_t]:
    return request(uri, HTTPMethods.PUT, headers=headers, body=body,
                   json=json, query=query, timeout=timeout,
                   blocksize=blocksize, datacls=datacls)


def delete(uri: str, 
           headers: Dict[str, str] = {},
           body: Optional[bytes] = None,
           json: Optional[Dict[str, Any]] = None,
           query: Dict[str, List[str]] = {},
           timeout: Optional[float] = None, 
           blocksize: int = 8192,
           datacls: Type[ResponseData_t] = BinaryApiData
           ) -> Response[ResponseData_t]:
    return request(uri, HTTPMethods.DELETE, headers=headers, body=body,
                   json=json, query=query, timeout=timeout,
                   blocksize=blocksize, datacls=datacls)


def head(uri: str,
         headers: Dict[str, str] = {},
         body: Optional[bytes] = None,
         json: Optional[Dict[str, Any]] = None,
         query: Dict[str, List[str]] = {},
         timeout: Optional[float] = None, 
         blocksize: int = 8192,
         datacls: Type[ResponseData_t] = BinaryApiData
         ) -> Response[ResponseData_t]:
    return request(uri, HTTPMethods.HEAD, headers=headers, body=body,
                   json=json, query=query, timeout=timeout,
                   blocksize=blocksize, datacls=datacls)


def options(uri: str, 
            headers: Dict[str, str] = {}, 
            body: Optional[bytes] = None,
            json: Optional[Dict[str, Any]] = None, 
            query: Dict[str, List[str]] = {},
            timeout: Optional[float] = None, 
            blocksize: int = 8192,
            datacls: Type[ResponseData_t] = BinaryApiData
            ) -> Response[ResponseData_t]:
    return request(uri, HTTPMethods.OPTIONS, headers=headers, body=body,
                   json=json, query=query, timeout=timeout,
                   blocksize=blocksize, datacls=datacls)




from __future__ import annotations
from http import client
from typing import Any, Dict, List, Optional, Type

from bamboo.api import BinaryApiData  
from bamboo.base import HTTPMethods
from bamboo.request import ResponseData_t, Schemes
from bamboo.request.request_form import get_http_request_form
from bamboo.request.response import Response


def request(uri: str, 
            method: str, 
            headers: Dict[str, str] = {},
            body: Optional[bytes] = None, 
            json: Optional[Dict[str, Any]] = None,
            query: Dict[str, List[str]] = {},
            timeout: Optional[float] = None, 
            blocksize: int = 8192,
            datacls: Type[ResponseData_t] = BinaryApiData
            ) -> Response[ResponseData_t]:
    form = get_http_request_form(Schemes.HTTP, uri, method, headers=headers,
                          body=body, json=json, query=query)
    conn = client.HTTPConnection(form.host, port=form.port,
                                 timeout=timeout, blocksize=blocksize)
    conn.request(form.method, form.uri, body=form.body, headers=form.headers)
    _res = conn.getresponse()
    return Response(conn, _res, datacls=datacls)


def get(uri: str, 
        headers: Dict[str, str] = {}, 
        body: Optional[bytes] = None,
        json: Optional[Dict[str, Any]] = None,
        query: Dict[str, List[str]] = {},
        timeout: Optional[float] = None, 
        blocksize: int = 8192,
        datacls: Type[ResponseData_t] = BinaryApiData
        ) -> Response[ResponseData_t]:
    return request(uri, HTTPMethods.GET, headers=headers, body=body,
                   json=json, query=query, timeout=timeout,
                   blocksize=blocksize, datacls=datacls)


def post(uri: str, 
        headers: Dict[str, str] = {},
        body: Optional[bytes] = None,
        json: Optional[Dict[str, Any]] = None,
        query: Dict[str, List[str]] = {},
        timeout: Optional[float] = None, 
        blocksize: int = 8192,
        datacls: Type[ResponseData_t] = BinaryApiData
        ) -> Response[ResponseData_t]:
    return request(uri, HTTPMethods.POST, headers=headers, body=body,
                   json=json, query=query, timeout=timeout,
                   blocksize=blocksize, datacls=datacls)


def put(uri: str, 
        headers: Dict[str, str] = {},
        body: Optional[bytes] = None,
        json: Optional[Dict[str, Any]] = None,
        query: Dict[str, List[str]] = {},
        timeout: Optional[float] = None, 
        blocksize: int = 8192,
        datacls: Type[ResponseData_t] = BinaryApiData
        ) -> Response[ResponseData_t]:
    return request(uri, HTTPMethods.PUT, headers=headers, body=body,
                   json=json, query=query, timeout=timeout,
                   blocksize=blocksize, datacls=datacls)


def delete(uri: str, 
           headers: Dict[str, str] = {},
           body: Optional[bytes] = None,
           json: Optional[Dict[str, Any]] = None,
           query: Dict[str, List[str]] = {},
           timeout: Optional[float] = None, 
           blocksize: int = 8192,
           datacls: Type[ResponseData_t] = BinaryApiData
           ) -> Response[ResponseData_t]:
    return request(uri, HTTPMethods.DELETE, headers=headers, body=body,
                   json=json, query=query, timeout=timeout,
                   blocksize=blocksize, datacls=datacls)


def head(uri: str,
         headers: Dict[str, str] = {},
         body: Optional[bytes] = None,
         json: Optional[Dict[str, Any]] = None,
         query: Dict[str, List[str]] = {},
         timeout: Optional[float] = None, 
         blocksize: int = 8192,
         datacls: Type[ResponseData_t] = BinaryApiData
         ) -> Response[ResponseData_t]:
    return request(uri, HTTPMethods.HEAD, headers=headers, body=body,
                   json=json, query=query, timeout=timeout,
                   blocksize=blocksize, datacls=datacls)


def options(uri: str, 
            headers: Dict[str, str] = {}, 
            body: Optional[bytes] = None,
            json: Optional[Dict[str, Any]] = None, 
            query: Dict[str, List[str]] = {},
            timeout: Optional[float] = None, 
            blocksize: int = 8192,
            datacls: Type[ResponseData_t] = BinaryApiData
            ) -> Response[ResponseData_t]:
    return request(uri, HTTPMethods.OPTIONS, headers=headers, body=body,
                   json=json, query=query, timeout=timeout,
                   blocksize=blocksize, datacls=datacls)




import random
import string
from typing import List
import unittest

from bamboo.app import App
from bamboo.endpoint import Endpoint
from bamboo.location import AsciiDigitLocation, AnyStringLocation, Uri_t
from bamboo.request import http
from bamboo.router import DuplicatedUriRegisteredError, Router
from bamboo.test import ServerForm, TestExecutor


def rand_string(k: int) -> str:
    return "".join(random.choices(string.ascii_letters + string.digits, k=k))


def rand_strings(num: int) -> List[str]:
    return [rand_string(random.randint(10, 20)) for _ in range(num)]


IDEAL_RESNPONSE = "Hello, World".encode()


# Template endpoints
class MockEndpoint(Endpoint):
    
    def do_GET(self) -> None:
        self.send_body(IDEAL_RESNPONSE)
        

class TestRouting(unittest.TestCase):
        
    def test_routing(self):
        # Make different uris
        URIS = list(set(rand_strings(100)))
        
        def client():
            for i, uri in enumerate(URIS):
                res = http.get(f"http://localhost:{8000 + i}/{uri}")
                self.assertEqual(res.body, IDEAL_RESNPONSE)
                
        forms = []
        for i, uri in enumerate(URIS):
            app = App()
            app.route(uri)(MockEndpoint)
            forms.append(ServerForm("", 8000 + i, app, "test_routing.log"))
        
        TestExecutor(*forms).exec(client)
        
    def _test_duplicated_uris(self, patterns: List[Uri_t]):
        router = Router()
        
        with self.assertRaises(DuplicatedUriRegisteredError) as err:
            for uri in patterns:
                router.register(uri, MockEndpoint)
                
        self.assertIsInstance(err.exception, DuplicatedUriRegisteredError)
            
    def test_duplicated_uris(self):
        PATTERNS_1 = [("test", "hoge", AnyStringLocation()),
                      ("test", "hoge", "image")]
        PATTERNS_2 = [(AnyStringLocation(), AsciiDigitLocation(4)),
                      ("hoge", "hoge")]
        PATTERNS_3 = [(AnyStringLocation(), "test", AnyStringLocation()),
                      ("hoge", AnyStringLocation(), "image")]
        
        self._test_duplicated_uris(PATTERNS_1)
        self._test_duplicated_uris(PATTERNS_2)
        self._test_duplicated_uris(PATTERNS_3)
            

if __name__ == "__main__":
    unittest.main()




import json
from typing import List, Optional, Tuple

from bamboo.base import (
    ContentType, HTTPStatus, ContentTypeHolder, AuthSchemes,
    DEFAULT_CONTENT_TYPE_PLAIN, DEFAULT_CONTENT_TYPE_JSON,
)
from bamboo.util.deco import class_property


class ErrInfoBase(ContentTypeHolder):
    """Base class of all error handlings.
    
    This class defines the attributes of all classes for error handling.
    
    Attributes
    ----------
    http_status : HTTPStatus
        HTTP status of the error
    message : Optional[str]
        short message for announcing error, default None
    """
    http_status: HTTPStatus = HTTPStatus.BAD_REQUEST
    
    # Optional short message for announcing error.
    # If None, will be set as default message.
    message: Optional[str] = None
    
    def __init_subclass__(cls) -> None:
        if cls.message is None:
           cls.message = cls.http_status.description
    
    def __init__(self, *args, **kwargs) -> None:
        pass
    
    @class_property
    def _content_type_(cls) -> ContentType:
        return DEFAULT_CONTENT_TYPE_PLAIN
    
    def get_headers(self) -> List[Tuple[str, str]]:
        """Publishes additional headers for error response.

        Returns
        -------
        List[Tuple[str, str]]
            Additional headers
        """
        return []
    
    def get_body(self) -> bytes:
        """Publishes response body for error response.

        Returns
        -------
        bytes
            Response body
        """
        return b""
    
    def get_all_form(self) -> Tuple[HTTPStatus, List[Tuple[str, str]], bytes]:
        stat = self.http_status
        headers = self.get_headers()
        body = self.get_body()
        return (stat, headers, body)
           
# ----------------------------------------------------------------------------

# Default errors    ----------------------------------------------------------

class DefaultNotFoundErrInfo(ErrInfoBase):
    
    http_status = HTTPStatus.NOT_FOUND


class DefaultDataFormatErrInfo(ErrInfoBase):
    
    http_status = HTTPStatus.UNSUPPORTED_MEDIA_TYPE


class DefaultHeaderNotFoundErrInfo(ErrInfoBase):
    
    http_status = HTTPStatus.BAD_REQUEST
    

class DefaultNotApplicableIpErrInfo(ErrInfoBase):
    
    http_status = HTTPStatus.FORBIDDEN


_WWW_AUTH_HEADER = "WWW-Authentication"


def get_auth_realm(scheme: str, msg: str) -> str:
    return f'{scheme} realm="{msg}"'


DEFAULT_REALM_MESSAGE = "SECRET AREA"


def get_default_realm(scheme: str) -> str:
    return get_auth_realm(scheme, DEFAULT_REALM_MESSAGE)


class DefaultAuthHeaderNotFoundErrInfo(ErrInfoBase):
    
    http_status = HTTPStatus.UNAUTHORIZED
    
    def __init__(self, scheme: str, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._scheme = scheme
        
    @property
    def scheme(self) -> str:
        return self._scheme
    
    def get_headers(self) -> List[Tuple[str, str]]:
        value = get_default_realm(self._scheme)
        return [(_WWW_AUTH_HEADER, value)]


DEFAULT_NOT_FOUND_ERROR = DefaultNotFoundErrInfo()
DEFUALT_INCORRECT_DATA_FORMAT_ERROR = DefaultDataFormatErrInfo()
DEFAULT_HEADER_NOT_FOUND_ERROR = DefaultHeaderNotFoundErrInfo()
DEFAULT_NOT_APPLICABLE_IP_ERROR = DefaultNotApplicableIpErrInfo()
DEFAULT_BASIC_AUTH_HEADER_NOT_FOUND_ERROR =\
    DefaultAuthHeaderNotFoundErrInfo(AuthSchemes.basic)
DEFAULT_DIGEST_AUTH_HEADER_NOT_FOUND_ERROR = \
    DefaultAuthHeaderNotFoundErrInfo(AuthSchemes.digest)
DEFAULT_BEARER_AUTH_HEADER_NOT_FOUND_ERROR = \
    DefaultAuthHeaderNotFoundErrInfo(AuthSchemes.bearer)

# ----------------------------------------------------------------------------

class ApiErrInfo(ErrInfoBase):
    """ErrInfo to handle API error.
    
    This ErrInfo has implemented method of 'get_body'. This class 
    emits Json data including error information defined by developer 
    when the error is sent.
    
    Attributes
    ----------
    http_status : HTTPStatus
        HTTP status of the error
    message : Optional[str]
        Short message for announcing error, by default None
    code : Optional[int]
        Error code for your API, by default None
    dev_message : Optional[str]
        Message to explain developers the error, by default None
    user_message : Optional[str]
        Message to explain end users the error, by default None
    info : Optional[str]
        Information about the error, by default None
    encoding : str
        Encoding to encode response body, by default 'UTF-8'
    """    
    code: Optional[int] = None
    dev_message: Optional[str] = None
    user_message: Optional[str] = None
    info: Optional[str] = None
    encoding: str = "UTF-8"
    
    @class_property
    def _content_type_(cls) -> ContentType:
        return DEFAULT_CONTENT_TYPE_JSON
    
    def get_body(self) -> Optional[bytes]:
        """Publishes response body for error response.

        Returns
        -------
        bytes
            Response body
        """
        body = {
            "code": self.code,
            "developerMessage": self.dev_message,
            "uesrMessage": self.user_message,
            "info": self.info
        }
        return json.dumps(body).encode(encoding=self.encoding)




from typing import List
import unittest

from bamboo.api import (
    XWWWFormUrlEncodedData, ValidationFailedError,
)
from bamboo.base import ContentType, MediaTypes


ideal_content_type = ContentType(MediaTypes.x_www_form_urlencoded, "utf-8")


class TestData(XWWWFormUrlEncodedData):
    
    name: str
    age: str
    email: str
    
    
data_raw = "name=hogehoge&email=hoge@hoge.com&age=18".encode()

data_key_not_included = "name=hogehoge&age=20".encode()

data_duplicated_key = "name=hogehoge&name=hogest&age=20&email=hoge@hoge.com"
data_duplicated_key = data_duplicated_key.encode()


class TestXWWWFromUrlEncodedData(unittest.TestCase):
    
    def test_data_raw(self):
        data = TestData(data_raw, ideal_content_type)
        self.assertEqual(data.name, "hogehoge")
        self.assertEqual(int(data.age), 18)
        self.assertEqual(data.email, "hoge@hoge.com")

    def test_data_key_not_included(self):
        with self.assertRaises(ValidationFailedError) as err:
            data = TestData(data_key_not_included, ideal_content_type)
        self.assertIsInstance(err.exception, ValidationFailedError)
        
    def test_data_duplicated_key(self):
        with self.assertRaises(ValidationFailedError) as err:
            data = TestData(data_duplicated_key, ideal_content_type)
        self.assertIsInstance(err.exception, ValidationFailedError)


if __name__ == "__main__":
    unittest.main()





import glob
import os
from typing import Any, Callable, Dict, List, Set, Type

from bamboo.app import App
from bamboo.endpoint import Endpoint
from bamboo.location import Location


def get_all_files(dir: str) -> Set[str]:
    paths = glob.glob(os.path.join(dir, "**"), recursive=True)
    return set(os.path.abspath(p) for p in paths if os.path.isfile(p))


class CommonNasEndpoint(Endpoint):
    
    def setup(self, file_path: str) -> None:
        self.file_path = file_path
    
    def do_GET(self) -> None:
        with open(self.file_path, "rb") as f:
            data = f.read()
            
        self.send_body(data)


class SimpleNasApp(App):
    
    def __init__(self, doc_root: str, 
                 endpoint: Type[CommonNasEndpoint]) -> None:
        super().__init__()  
        
        if doc_root[-1] == "/":
            doc_root = doc_root[:-1]
        
        self._doc_root = os.path.abspath(doc_root)
        self._endpoint = endpoint
        self._files = get_all_files(self._doc_root)
        self._restricted_files: Set[str] = set()
                
    def __call__(self, environ: Dict[str, Any], 
                 start_response: Callable) -> List[bytes]:
        method = environ.get("REQUEST_METHOD").upper()
        path = self._doc_root + environ.get("PATH_INFO")
        if path in self._restricted_files:
            return [self.send_404(start_response)]
            
        if path in self._files:
            endpoint = self._endpoint(environ, path)
            callback = self._endpoint._get_response_method(method)
            if callback is None:
                return [self.send_404(start_response)]
            
            callback(endpoint)
            start_response(endpoint._status.value, endpoint._headers)
            return [endpoint._res_body]
        
        # If no resources found, try to call other callbacks.
        return super().__call__(environ, start_response)

    def restrict_uri(self, *locs: Location) -> None:
        path = os.path.join(self._doc_root, "/".join(locs))
        self._restricted_files.add(path)





from typing import (
    Any, Callable, Generic, Optional, Type, TypeVar, Union,
)


def joint_funcs(*func: Callable[[], None]) -> Callable[[], None]:
    def joint():
        for f in func:
            f()
    return joint


def joint_methods(*methods: Callable[[Any], None]) -> Callable[[Any], None]:
    def joint(obj: Any):
        for m in methods:
            m(obj)
    return joint


Object = TypeVar("Object")
ReturnGetter = TypeVar("ReturnGetter")


class cached_property(Generic[Object, ReturnGetter]):
    
    def __init__(self, 
                 fget: Callable[[Object], ReturnGetter]) -> None:
        self.__doc__ = getattr(fget, "__doc__")
        self._fget = fget
        
    def __get__(self, obj: Union[Object, None], 
                clazz: Optional[Type[Object]] = None) -> ReturnGetter:
        if obj is None:
            return self
        if self._fget is not None:
            value = self._fget(obj)
            obj.__dict__[self._fget.__name__] = value
            return value
        raise AttributeError(
            "'getter' has not been set yet."
        )
    
    def getter(self, fget: Callable[[Object], ReturnGetter]):
        self._fget = fget
        return self
    
    
class class_property(Generic[Object, ReturnGetter]):
    
    def __init__(self,
                 fget: Callable[[Object], ReturnGetter]) -> None:
        self.__doc__ = getattr(fget, "__doc__")
        self._fget = fget
        
    def __get__(self, obj: Union[Object, None], 
                clazz: Optional[Type[Object]] = None) -> ReturnGetter:
        if clazz is None:
            clazz = type(obj)
            
        if self._fget is not None:
            return self._fget(clazz)
        raise AttributeError(
            "'getter' has not been set yet."
        )
            
    def getter(self, fget: Callable[[Object], ReturnGetter]):
        self._fget = fget
        return self




from __future__ import annotations
from typing import (
    Any, Callable, Dict, List, Optional, Tuple, Type, Union,
)

from bamboo.endpoint import Endpoint, _get_bamboo_attr
from bamboo.error import DEFAULT_NOT_FOUND_ERROR
from bamboo.location import Location_t, Uri_t
from bamboo.router import Router


ATTR_VERSION = _get_bamboo_attr("version")
Version_t = Tuple[int]


class VersionConfig:
    """Operator class for version of `Endpoint`.
    
    This class can be used to get and set version of `Endpoint` safely.
    """
    
    def __init__(self, endpoint: Type[Endpoint]) -> None:
        """
        Parameters
        ----------
        endpoint : Type[Endpoint]
            `Endpoint` whose version is to be manipulated
        """
        self._endpoint_class = endpoint
        
    def set(self, app: App, version: Union[int, Tuple[int], None] = None,
            force: bool = False) -> None:
        """Set version of `Endpoint`.

        Parameters
        ----------
        app : App
            Application including the internal `Endpoint`
        version : Union[int, Tuple[int], None], optional
            Version to be set, by default `None`
        force : bool, optional
            If forcing to set the `version`, by default `False`

        Raises
        ------
        ValueError
            Raised if version of the `Endpoint` has already been set.
        """
        if not hasattr(self._endpoint_class, ATTR_VERSION):
            setattr(self._endpoint_class, ATTR_VERSION, {})
        
        registered = getattr(self._endpoint_class, ATTR_VERSION)
        current_version = registered.get(app)
        if current_version and not force:
            raise ValueError(
                f"{self._endpoint_class.__name__} already has own version."
                "If you want to overwirte it, set param 'force' True.")
        
        # Format to fit the type Version_t
        if version is None:
            version = ()
        if isinstance(version, int):
            version = (version,)
            
        registered[app] = version
        
    def get(self, app: App) -> Optional[Version_t]:
        """Retrieve version belonging to specified `app`.

        Parameters
        ----------
        app : App
            Application including the internal `Endpoint`

        Returns
        -------
        Optional[Version_t]
            Version set to `Endpoint`, if not set yet, then `None`
        """
        if hasattr(self._endpoint_class, ATTR_VERSION):
            registered = getattr(self._endpoint_class, ATTR_VERSION)
            return registered.get(app)
        return None
    
    def get_all(self) -> List[Tuple[App, Version_t]]:
        """Retrieve versions belonging to all `App` objects.

        Returns
        -------
        List[Tuple[App, Version_t]]
            `list` of `tuple`s of `App` objects and their versions
        """
        if hasattr(self._endpoint_class, ATTR_VERSION):
            registered = getattr(self._endpoint_class, ATTR_VERSION)
            return [(app, version) for app, version in registered.items()]
        return []


ATTR_PARCEL = _get_bamboo_attr("parcel")
Parcel_t = Tuple[Any, ...]


class ParcelConfig:
    """Operator class for parcel of `Endpoint`.
    
    This class can be used to get and set parcel of `Endpoint` safely.
    """
    
    def __init__(self, endpoint: Type[Endpoint]) -> None:
        """
        Parameters
        ----------
        endpoint : Type[Endpoint]
            `Endpoint` whose parcel is to be manipulated
        """
        self._endpoint_class = endpoint
        
    def set(self, app: App, parcel: Parcel_t) -> None:
        """Ser parcel of `Endpoint`

        Parameters
        ----------
        app : App
            Application including the internal `Endpoint`
        parcel : Parcel_t
            Parcel to be set
        """
        if not hasattr(self._endpoint_class, ATTR_PARCEL):
            setattr(self._endpoint_class, ATTR_PARCEL, {})
        
        registered = getattr(self._endpoint_class, ATTR_PARCEL)
        registered[app] = parcel
        
    def get(self, app: App) -> Parcel_t:
        """Retrieve parcel belonging to specified `app`/

        Parameters
        ----------
        app : App
            Application including the internal `Endpoint`

        Returns
        -------
        Parcel_t
            Parcel set to `Endpoint`, if not set yet, then `None`
        """
        if hasattr(self._endpoint_class, ATTR_PARCEL):
            registered = getattr(self._endpoint_class, ATTR_PARCEL)
            return registered.get(app)
        return ()
        
    def get_all(self) -> List[Tuple[App, Parcel_t]]:
        """Retrieve parcels belonging to all `App` objects.

        Returns
        -------
        List[Tuple[App, Parcel_t]]
            `list` of `tuple`s of `App` objects and their parcels
        """
        if hasattr(self._endpoint_class, ATTR_PARCEL):
            registered = getattr(self._endpoint_class, ATTR_PARCEL)
            return [(app, parcel) for app, parcel in registered.items()]
        return []


class App:
    """Application objects compliant with the WSGI.

    Attributes
    ----------
    TAG_VERSION : str
        Prefix of version
    """
    
    TAG_VERSION = "v"
    
    def __init__(self, is_version_inserted: bool = True) -> None:
        """
        Parameters
        ----------
        is_version_inserted : bool, optional
            If version is inserted at the head of paths of URIs, 
            by default `True`
        """
        self._router = Router()
        self._is_version_isnerted = is_version_inserted
    
    def __call__(self, environ: Dict[str, Any],
                 start_response: Callable) -> List[bytes]:
        method = environ.get("REQUEST_METHOD").upper()
        path = environ.get("PATH_INFO")
        
        flexible_locs, endpoint_class = self._router.validate(path)
        if endpoint_class is None:
            return [self._send_404(start_response)]
        
        parcel_config = ParcelConfig(endpoint_class)
        parcel = parcel_config.get(self)
        
        endpoint = endpoint_class(environ, flexible_locs, *parcel)
        callback = endpoint_class._get_response_method(method)
        if callback is None:
            return [self._send_404(start_response)]
        
        callback(endpoint)
        start_response(endpoint._status.value, endpoint._headers)
        return [endpoint._res_body]
    
    @staticmethod
    def _send_404(start_response: Callable) -> bytes:
        """Send `404` error code, i.e. `Resource Not Found` error.

        Parameters
        ----------
        start_response : Callable
            `start_response` callbale given from the WSGI application

        Returns
        -------
        bytes
            Response body
        """
        stat, headers, res_body = DEFAULT_NOT_FOUND_ERROR.get_all_form()
        start_response(stat.value, headers)
        return res_body
    
    def seach_uris(self, endpoint: Type[Endpoint]) -> List[Uri_t]:
        """Retrieve all URI patterns of `Endpoint`.

        Parameters
        ----------
        endpoint : Type[Endpoint]
            `Endpoint` whose URIs to be searched

        Returns
        -------
        List[Uri_t]
            Result of searching
        """
        return self._router.search_uris(endpoint)
    
    def route(self, *locs: Location_t,
              version: Union[int, Tuple[int], None] = None
              ) -> Callable[[Type[Endpoint]], Type[Endpoint]]:
        """Register combination of URI and `Endpoint` for routing.

        Parameters
        ----------
        version : Union[int, Tuple[int], None], optional
            Version of the `Endpoint`, by default None

        Returns
        -------
        Callable[[Type[Endpoint]], Type[Endpoint]]
            Decorator to add combination of URI and `Endpoint`
            
        Examples
        --------
        ```
        app = App()
        
        # set path of URI as `test/data/image` and version as 1
        @app.route("test", "data", "image", version=1)
        class MockEndpoint(Endpoint):
            
            def do_GET(self) -> None:
                # Do something...
        ```
        """
        def register_endpoint(endpoint: Type[Endpoint]) -> Type[Endpoint]:            
            # version setting
            ver_config = VersionConfig(endpoint)
            ver_config.set(self, version)
            
            # router setting
            _version = ver_config.get(self)
            if self._is_version_isnerted and len(_version):
                _version = ver_config.get(self)
                if len(_version):
                    uri_list = [(f"{self.TAG_VERSION}{ver_num}",) + locs 
                                for ver_num in _version]
                    for uri_version_included in uri_list:
                        self._router.register(uri_version_included, endpoint)
            else:
                self._router.register(locs, endpoint)

            return endpoint
        return register_endpoint
    
    def set_parcel(self, endpoint: Type[Endpoint], *parcel: Any) -> None:
        """Set parcel to an endpoint.

        Parameters
        ----------
        endpoint : Type[Endpoint]
            `Endpoint` the `parcel` to be set
        *parcel : Any
            Pacel to be given to the `Endpoint`
        """
        parcel_config = ParcelConfig(endpoint)
        parcel_config.set(self, parcel)





import base64
from typing import Dict, List, Union


def encode_binary(data: Union[bytes, bytearray]) -> str:
    return base64.b64encode(data).decode()


def decode2binary(data: str) -> bytes:
    return base64.b64decode(data.encode())


def encode_base64_string(data: str) -> str:
    return encode_binary(data.encode())


def unparse_qs(query: Dict[str, List[str]]) -> str:
    params = []
    for key, vals in query.items():
        repr_vals = ",".join(vals)
        params.append("=".join((key, repr_vals)))
    return "&".join(params)




from typing import Dict, List, Optional, Tuple, Type

from bamboo.endpoint import Endpoint
from bamboo.location import (
    FlexibleLocation, Uri_t, is_flexible_uri, is_duplicated_uri,
)


# Type definition
HTTPMethod_t = str
Uri2Endpoints_t = Dict[Uri_t, Type[Endpoint]]


class DuplicatedUriRegisteredError(Exception):
    """Raised if duplicated URI is registered."""
    pass


class Router:
    """Operator of routing request to `Endpoint` by URI.
    """
    
    def __init__(self) -> None:
        self.uri2endpoint: Uri2Endpoints_t = {}
        self.uris_flexible: List[Uri_t] = []
    
    def register(self, uri: Uri_t, endpoint: Type[Endpoint]) -> None:
        """Register combination of URI and `Endpoint`.

        Parameters
        ----------
        uri : Uri_t
            URI pattern of the `Endpoint`
        endpoint : Type[Endpoint]
            `Endpoint` class to be registered

        Raises
        ------
        DuplicatedUriRegisteredError
            Raised if given URI pattern matches one already registered
        """
        for uri_registered in self.uri2endpoint.keys():
            if is_duplicated_uri(uri_registered, uri):
                raise DuplicatedUriRegisteredError(
                    "Duplicated URIs were detected.")
        
        if is_flexible_uri(uri):
            self.uris_flexible.append(uri)
        self.uri2endpoint[uri] = endpoint
    
    def validate(self, uri: str
                 ) -> Tuple[Tuple[str, ...], Optional[Type[Endpoint]]]:
        """Validate specified `uri` and retrieved `Endpoint`.

        Parameters
        ----------
        uri : str
            Path of URI

        Returns
        -------
        Tuple[Tuple[str, ...], Optional[Type[Endpoint]]]
            Pair of values of flexible locations and `Endpoint` if specified 
            `uri` is valid
            
        Notes
        -----
        This method returns pair of tuple of flexible locations specified as 
        parts of URI pattern linked to `Endpoint` and its `Endpoint`. If any 
        flexible locations are not included in the URI pattern, then empty 
        tuple will be returned as a sequence of flexible locations, so if 
        URI patterns is configured with only static locations, you will get 
        the empty tuple. 
        
        If invalid URI pattern is come, then also empty tuple will be return 
        as sequence of flexible locations and `None` as `Endpoint`, or 
        `((), None)`.
        """
        uri = tuple(uri[1:].split("/"))
        endpoint = self.uri2endpoint.get(uri)
        if endpoint:
            return ((), endpoint)
        
        depth = len(uri)
        for flexible in self.uris_flexible:
            if len(flexible) != depth:
                continue
            
            flexibles_received = []
            
            # Judging each locations
            for loc_req, loc_flex in zip(uri, flexible):
                if loc_req == loc_flex:
                    continue
                if isinstance(loc_flex, FlexibleLocation):
                    if not loc_flex.is_valid(loc_req):
                        break
                    
                    flexibles_received.append(loc_req)
            else:
                # Correct case
                endpoint = self.uri2endpoint.get(flexible)
                return (tuple(flexibles_received), endpoint)
        
        # Could not find it
        return ((), None)
    
    def search_uris(self, endpoint: Type[Endpoint]) -> List[Uri_t]:
        """Search URI patterns of specified `endpoint`.

        Parameters
        ----------
        endpoint : Type[Endpoint]
            `Endpoint` whose URI patterns to be retrieved

        Returns
        -------
        List[Uri_t]
            Result of searching
        """
        return [uri for uri, point in self.uri2endpoint.items() 
                if point is endpoint]




from __future__ import annotations
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
import re
from typing import Optional
from enum import Enum

from bamboo.util.deco import class_property


class _HTTPMethods:
    """Singleton class to iterate HTTP methods."""
    
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    HEAD = "HEAD"
    OPTIONS = "OPTIONS"
    
    __methods = set((
        GET, POST, PUT, DELETE, HEAD, OPTIONS
    ))
    
    __instance = None
    
    def __new__(cls) -> _HTTPMethods:
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance
    
    def __iter__(self):
        return iter(self.__methods)
    
    def __contains__(self, item: str):
        return item in self.__methods
    

HTTPMethods = _HTTPMethods()


class _MediaTypes:
    """Singleton class to iterate media types for communication."""
    
    plain = "text/plain"
    html = "text/html"
    xml = "application/xml"
    css = "text/css"
    javascript = "application/javascript"
    json = "application/json"
    x_www_form_urlencoded = "application/x-www-form-urlencoded"
    rss = "application/rss+xml"
    atom = "application/atom+xml"
    binary = "application/octet-stream"
    zip = "application/zip"
    jpeg = "image/jpeg"
    png = "image/png"
    svg = "image/svg+xml"
    multi_form = "multipart/form-data"
    mp4 = "video/mp4"
    excel = "application/vnd.ms-excel"
    
    __types = set((
        plain, html, xml, css, javascript, json, rss, atom,
        binary, zip, jpeg, png, svg, multi_form, mp4, excel,
    ))
    
    __instance = None
    
    def __new__(cls) -> _MediaTypes:
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance
    
    def __iter__(self):
        return iter(self.__types)
    
    def __contains__(self, item: str):
        return item in self.__types
    
    
MediaTypes = _MediaTypes()


class _AuthSchemes:
    """Singleton class to iterate authentication schemes of 
    the Authorization header.
    """
    
    basic = "Basic"
    bearer = "Bearer"
    digest = "Digest"
    
    __schemes = set((basic, bearer, digest))
    
    __instance = None
    
    def __new__(cls) -> _AuthSchemes:
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance
    
    def __iter__(self):
        return iter(self.__schemes)
    
    def __contains__(self, item: str):
        return item in self.__schemes
    
    
AuthSchemes = _AuthSchemes()


@dataclass
class ContentType:
    """`dataclass` for describing value of `Content-Type` header.
    """
    
    media_type: Optional[str] = None
    charset: Optional[str] = None
    boundary: Optional[str] = None
    
    @classmethod
    def parse(cls, raw: str) -> ContentType:
        """Parse and make new instance of this class.

        Parameters
        ----------
        raw : str
            Value of `Content-Type` header

        Returns
        -------
        ContentType
            New instance of this class based on the `raw` data
        """
        raw = re.split(";\s*", raw)
        result = cls()
        
        media_type = raw[0]
        if len(media_type):
            result.media_type = media_type
            
        for directive, val in [item.split("=") for item in raw[1:]]:
            if directive == "charset":
                result.charset = val.lower()
            elif directive == "boundary":
                result.boundary = val
                
        return result
    
    
class ContentTypeHolder(metaclass=ABCMeta):
    """Abstract class with properties about `Content-Type`.
    """
    
    @class_property
    @abstractmethod
    def _content_type_(cls) -> ContentType:
        pass

    
# NOTE
#   Value of the Content-Type header by default. This variables can be 
#   used if Content-Type of response is not specified.
DEFAULT_CONTENT_TYPE_PLAIN = ContentType(MediaTypes.plain)
DEFAULT_CONTENT_TYPE_JSON = ContentType(MediaTypes.json, "UTF-8")


class HTTPStatus(str, Enum):
    """Enum which describes HTTP status.

    Each class variables of the class have two main attributes of 
    'value' and 'description' as strings. 'value' means the enum value 
    and describes HTTP status for WSGI. 'description' is additional 
    attributes of the Enum, and describes short message of 
    the corresponding value.
    """

    def __new__(cls, value: str, description: str = ""):
        obj = str.__new__(cls, value)
        obj._value_ = value
        obj.description = description
        return obj

    CONTINUE = (
        "100 Continue",
        "Request received, please continue"
    )
    SWITCHING_PROTOCOLS = (
        "101 Switching Protocols",
        "Switching to new protocol; obey Upgrade header"
    )
    PROCESSING = ("102 Processing", "")

    # success
    OK  = (
        "200 OK", 
        "Request fulfilled, document follows"
    )
    CREATED = (
        "201 Created", 
        "Document created, URL follows"
    )
    ACCEPTED = (
        "202 Accepted",
        "Request accepted, processing continues off-line"
    )
    NON_AUTHORITATIVE_INFORMATION = (
        "203 Non-Authoritative Information",
        "Request fulfilled from cache"
    )
    NO_CONTENT = (
        "204 No Content", 
        "Request fulfilled, nothing follows"
    )
    RESET_CONTENT = (
        "205 Reset Content",
        "Clear input form for further input"
    )
    PARTIAL_CONTENT = (
        "206 Partial Content",
        "Partial content follows"
    )
    MULTI_STATUS = ("207 Multi-Status", "")
    ALREADY_REPORTED = ("208 Already Reported", "")
    IM_USED = ("226 IM Used", "")

    # redirection
    MULTIPLE_CHOICES = (
        "300 Multiple Choices",
        "Object has several resources -- see URI list"
    )
    MOVED_PERMANENTLY = (
        "301 Moved Permanently",
        "Object moved permanently -- see URI list"
    )
    FOUND = (
        "302 Found",
        "Object moved temporarily -- see URI list"
    )
    SEE_OTHER = (
        "303 See Other", 
        "Object moved -- see Method and URL list"
    )
    NOT_MODIFIED = (
        "304 Not Modified",
        "Document has not changed since given time"
    )
    USE_PROXY = (
        "305 Use Proxy",
        "You must use proxy specified in Location to access this resource"
    )
    TEMPORARY_REDIRECT = (
        "307 Temporary Redirect",
        "Object moved temporarily -- see URI list"
    )
    PERMANENT_REDIRECT = (
        "308 Permanent Redirect",
        "Object moved permanently -- see URI list"
    )

    # client error
    BAD_REQUEST = (
        "400 Bad Request",
        "Bad request syntax or unsupported method"
    )
    UNAUTHORIZED = (
        "401 Unauthorized",
        "No permission -- see authorization schemes"
    )
    PAYMENT_REQUIRED = (
        "402 Payment Required",
        "No payment -- see charging schemes"
    )
    FORBIDDEN = (
        "403 Forbidden",
        "Request forbidden -- authorization will not help"
    )
    NOT_FOUND = (
        "404 Not Found",
        "Nothing matches the given URI"
    )
    METHOD_NOT_ALLOWED = (
        "405 Method Not Allowed",
        "Specified method is invalid for this resource"
    )
    NOT_ACCEPTABLE = (
        "406 Not Acceptable",
        "URI not available in preferred format"
    )
    PROXY_AUTHENTICATION_REQUIRED = (
        "407 Proxy Authentication Required",
        "You must authenticate with this proxy before proceeding"
    )
    REQUEST_TIMEOUT = (
        "408 Request Timeout",
        "Request timed out; try again later"
    )
    CONFLICT = (
        "409 Conflict", 
        "Request conflict"
    )
    GONE = (
        "410 Gone",
        "URI no longer exists and has been permanently removed"
    )
    LENGTH_REQUIRED = (
        "411 Length Required",
        "Client must specify Content-Length"
    )
    PRECONDITION_FAILED = (
        "412 Precondition Failed",
        "Precondition in headers is false"
    )
    REQUEST_ENTITY_TOO_LARGE = (
        "413 Request Entity Too Large",
        "Entity is too large"
    )
    REQUEST_URI_TOO_LONG = (
        "414 Request-URI Too Long",
        "URI is too long"
    )
    UNSUPPORTED_MEDIA_TYPE = (
        "415 Unsupported Media Type",
        "Entity body in unsupported format"
    )
    REQUESTED_RANGE_NOT_SATISFIABLE = (
        "416 Requested Range Not Satisfiable",
        "Cannot satisfy request range"
    )
    EXPECTATION_FAILED = (
        "417 Expectation Failed",
        "Expect condition could not be satisfied"
    )
    MISDIRECTED_REQUEST = (
        "421 Misdirected Request",
        "Server is not able to produce a response"
    )
    UNPROCESSABLE_ENTITY = ("422 Unprocessable Entity", "")
    LOCKED = ("423 Locked", "")
    FAILED_DEPENDENCY = ("424 Failed Dependency", "")
    UPGRADE_REQUIRED = ("426 Upgrade Required", "")
    PRECONDITION_REQUIRED = (
        "428 Precondition Required",
        "The origin server requires the request to be conditional"
    )
    TOO_MANY_REQUESTS = (
        "429 Too Many Requests",
        "The user has sent too many requests in a given amount of time"
        "('rate limiting')"
    )
    REQUEST_HEADER_FIELDS_TOO_LARGE = (
        "431 Request Header Fields Too Large",
        "The server is unwilling to process the request because its "
        "header fields are too large"
    )

    # server errors
    INTERNAL_SERVER_ERROR = (
        "500 Internal Server Error",
        "Server got itself in trouble"
    )
    NOT_IMPLEMENTED = (
        "501 Not Implemented",
        "Server does not support this operation"
    )
    BAD_GATEWAY = (
        "502 Bad Gateway",
        "Invalid responses from another server/proxy"
    )
    SERVICE_UNAVAILABLE = (
        "503 Service Unavailable",
        "The server cannot process the request due to a high load"
    )
    GATEWAY_TIMEOUT = (
        "504 Gateway Timeout",
        "The gateway server did not receive a timely response"
    )
    HTTP_VERSION_NOT_SUPPORTED = (
        "505 HTTP Version Not Supported",
        "Cannot fulfill request"
    )
    VARIANT_ALSO_NEGOTIATES = ("506 Variant Also Negotiates", "")
    INSUFFICIENT_STORAGE = ("507 Insufficient Storage", "")
    LOOP_DETECTED = ("508 Loop Detected", "")
    NOT_EXTENDED = ("510 Not Extended", "")
    NETWORK_AUTHENTICATION_REQUIRED = (
        "511 Network Authentication Required",
        "The client needs to authenticate to gain network access"
    )




import random
import unittest

import bamboo
from bamboo.request import http
from bamboo.util.string import rand_string
from bamboo.util.time import get_datetime_rfc822


app = bamboo.App()


class InfoResponse(bamboo.JsonApiData):
    
    name: str
    datetime: str


@app.route(bamboo.AnyStringLocation(), "info")
class InfoEndpoint(bamboo.Endpoint):
    
    @bamboo.data_format(input=None, output=InfoResponse)
    def do_GET(self) -> None:
        name, = self.flexible_locs
        
        body = {"name": name, "datetime": get_datetime_rfc822()}
        self.send_json(body)
        
        
class CalculationResult(bamboo.JsonApiData):

    result: int
        
        
@app.route(bamboo.AsciiDigitLocation(10), 
           bamboo.AsciiDigitLocation(10),
           "multiply")
class MultiplyEndpoint(bamboo.Endpoint):
    
    @bamboo.data_format(input=None, output=CalculationResult)
    def do_GET(self) -> None:
        num_1, num_2 = self.flexible_locs
        
        body = {"result": int(num_1) * int(num_2)}
        self.send_json(body)
        
        
class FlexibleLocsTest(unittest.TestCase):
    
    def setUp(self) -> None:
        form = bamboo.ServerForm("", 8000, app, "test_flexible_locs.log")
        self.executor = bamboo.TestExecutor(form)
        self.executor.start_serve()
    
    def tearDown(self) -> None:
        self.executor.close()
        
    def test_info(self):
        names = [rand_string(10) for _ in range(100)]
        
        for name in names:
            res = http.get(f"http://localhost:8000/{name}/info",
                           datacls=InfoResponse)
            if res.ok:
                body = res.attach()
                self.assertEqual(body.name, name)
            else:
                print(f"Request failed. Status code: {res.status}")
                
    def test_multiply(self):
        get_num = lambda: random.randint(10**9, 10**10 - 1)
        num_pairs = [(get_num(), get_num()) for _ in range(100)]
        
        for num_1, num_2 in num_pairs:
            res = http.get(f"http://localhost:8000/{num_1}/{num_2}/multiply",
                           datacls=CalculationResult)
            if res.ok:
                body = res.attach()
                self.assertEqual(body.result, num_1 * num_2)
            else:
                print(f"Request failed. Status code: {res.status}")
                
                
if __name__ == "__main__":
    unittest.main()




__version__ = "0.4.1"

from bamboo.api import (
    ApiData, ValidationFailedError, BinaryApiData, InvalidAnnotationError,
    JsonApiData, XWWWFormUrlEncodedData,
)

from bamboo.app import (
    App, ParcelConfig, ATTR_PARCEL, VersionConfig, ATTR_VERSION,
)

from bamboo.base import (
    ContentType, ContentTypeHolder, MediaTypes, HTTPMethods, HTTPStatus,
)

from bamboo.endpoint import (
    StatusCodeAlreadySetError, Endpoint, ATTR_DATA_FORMAT, DataFormatInfo,
    get_data_format_info, data_format, ATTR_HEADERS_REQUIRED, 
    RequiredHeaderInfo, get_required_header_info, has_header_of,
    ATTR_CLIENT_RESTRICTED, get_restricted_ip_info, restricts_client,
    ATTR_ERRORS, get_errors_info, may_occur, _get_bamboo_attr,
)

from bamboo.error import (
    ErrInfoBase, ApiErrInfo, DEFAULT_NOT_FOUND_ERROR, 
    DEFUALT_INCORRECT_DATA_FORMAT_ERROR, DEFAULT_HEADER_NOT_FOUND_ERROR,
    DEFAULT_NOT_APPLICABLE_IP_ERROR,
)

from bamboo.location import (
    FlexibleLocation, is_flexible_uri, is_duplicated_uri,
    AsciiDigitLocation, AnyStringLocation,
)

from bamboo.router import DuplicatedUriRegisteredError, Router

from bamboo.test import ServerForm, TestExecutor





import unittest

from bamboo import App, Endpoint, data_format
from bamboo.api import JsonApiData
from bamboo.base import ContentType, MediaTypes
from bamboo.request import http
from bamboo.test import  ServerForm, TestExecutor
from bamboo.util.time import get_datetime_rfc822


class InfoResponse(JsonApiData):
    
    server_name: str
    current_time: str


class MockInfoEndpoint(Endpoint):
    
    NAME_SERVER = "Mocker"
    
    @data_format(input=None, output=InfoResponse)
    def do_GET(self) -> None:
        body = {
            "server_name": self.NAME_SERVER,
            "current_time": get_datetime_rfc822()
        }
        self.send_json(body)


class MockImageEndpoint(Endpoint):
    
    PATH_IMAGE = "elephant.jpg"

    def do_GET(self) -> None:
        with open(self.PATH_IMAGE, "rb") as f:
            image = f.read()
            
        self.send_body(image, ContentType(MediaTypes.jpeg))
        
        
class TestHTTPRequest(unittest.TestCase):
    
    def setUp(self) -> None:
        self.urls_info = [f"http://localhost:800{i}/mock/info" for i in range(3)]
        self.urls_image = [f"http://localhost:800{i}/mock/image" for i in range(3)]
        self.path_image_ideal = "elephant.jpg"
        
        app_1 = App()
        app_2 = App()
        app_3 = App()

        for app in (app_1, app_2, app_3):
            app.route("mock", "info")(MockInfoEndpoint)
            app.route("mock", "image")(MockImageEndpoint)
            
        form_1 = ServerForm("", 8000, app_1, "serverlog_1.log")
        form_2 = ServerForm("", 8001, app_2, "serverlog_2.log")
        form_3 = ServerForm("", 8002, app_3, "serverlog_3.log")
        self.executor = TestExecutor(form_1, form_2, form_3).start_serve()

    def tearDown(self) -> None:
        self.executor.close()

    def test_get_info(self):
        for url in self.urls_info:
            res = http.request(url, "GET", datacls=InfoResponse)
            data = res.attach()
            
            self.assertTrue(isinstance(data, InfoResponse))
            self.assertEqual(data.server_name, "Mocker")
            print(f"Response of 'current_time' : {data.current_time}")
        
    def test_get_image(self):
        with open(self.path_image_ideal, "rb") as f:
            image_ideal = f.read()
        
        for url in self.urls_image:
            res = http.request(url, "GET")
            data = res.body
            self.assertEqual(image_ideal, data)


if __name__ == "__main__":
    unittest.main()




from __future__ import annotations
from abc import ABCMeta, abstractmethod
import json
from typing import (
    Any, Dict, Sequence, Type, TypeVar, Union, get_args,
    get_origin, get_type_hints,
)
from urllib.parse import parse_qs

from bamboo.base import MediaTypes, ContentTypeHolder, ContentType
from bamboo.util.deco import class_property


class ApiData(ContentTypeHolder, metaclass=ABCMeta):
    """Base class to describe input/output data format on Endpoints as APIs.
    
    Subclasses of this class can be used an argument of `data_format` 
    decorator for callbacks on Endpoints. If input/output parameters of 
    the `data_format` are specified, the decorator validates if raw data, 
    typically a `bytes` object, has expected data format. 
    
    Subclasses of this class can define their own data formats in `__init__` 
    methods. Developers should implement `__init__` methods of the subclasses 
    such that each objects has data with expected formats by implementors. 
    
    Notes
    -----
    This class is an abstract class. So, don't initilize it and specify it 
    as an argument of the `data_format` decorator.
    
    Subclasses of this class should validate if raw data given to `__init__` 
    methods has expected formats. If the validation failes, the classes MUST 
    raise `ValidataionFailedError` to announce the failure to the 
    `data_format` decorator.
    """

    @abstractmethod
    def __init__(self, raw: bytes, content_type: ContentType) -> None:
        """
        Parameters
        ----------
        raw : bytes
            Raw data to be validated
        content_type : ContentType
            Values of `Content-Type` header
        """
        pass
    
    @class_property
    def _content_type_(cls) -> str:
        return MediaTypes.plain


class ValidationFailedError(Exception):
    """Raised if type validation of data failes."""
    pass


class BinaryApiData(ApiData):
    """API data with no format.
    
    This class can be used to describe raw data with no data format. So, 
    any received data from clients is acceptable on the class.
    
    Attributes
    ----------
    raw : bytes
        Property of raw data
        
    Examples
    --------
    ```
    class MockEndpoint(Endpoint):
    
        @data_format(input=BinaryApiData, output=None)
        def do_GET(self, rec_body: BinaryApiData) -> None:
            # get raw data of request body
            raw_data = rec_body.raw
    ```
    """
    
    def __init__(self, raw: bytes, content_type: ContentType) -> None:
        """
        Parameters
        ----------
        raw : bytes
            Raw data to be validated
        content_type : ContentType
            Values of `Content-Type` header
            
        Notes
        -----
        In objects of this class, `content_type` is not used even if any 
        `content_type` is specified.
        """
        super().__init__(raw, content_type)
        
        if not isinstance(raw, bytes):
            raise ValidationFailedError(
                f"'raw' must be a 'bytes', but was {raw.__class__.__name__}.")
        self._raw = raw
        
    @property
    def raw(self) -> bytes:
        return self._raw

        
class InvalidAnnotationError(Exception):
    """Raised if a type not convertable to JSON is included of type hints."""
    pass


ListJsonable_t = TypeVar("ListJsonable_t")
    
    
class JsonApiDataBuilder:
    """Builder for `JsonApiData`.
    
    This utility class has some class methods for validating JSON data and 
    building `JsonApiData` objects from raw data.
    
    Attributes
    ----------
    NoneType : type
        The type of None
    TYPES_ARGS : Tuple[type]
        Tuple of types acceptable to `JsonApiData` with no arguments
    TYPES_ARGS_SET : Set[type]
        Set of types acceptable to `JsonApiData` with no arguments
    TYPES_ORIGIN : Tuple[type]
        Tuple of types acceptable to `JsonApiData` with arguments
    """
    
    NoneType = type(None)
    TYPES_ARGS = (int, float, str, bool, NoneType)
    TYPES_ARGS_SET = set(TYPES_ARGS)
    TYPES_ORIGIN = (list, Union)
    
    @classmethod
    def check_annotations(cls, *types: Type) -> None:
        """Check if types are convertable to JSON data.
        
        Parameters
        ----------
        *types : Type
            Classes to be checked
            
        Raises
        ------
        InvalidAnnotationError
            Raised if a type not convertable to JSON data is detected
        """  
        def checker(types: Sequence[Type]) -> None:
            for type_ in types:
                origin = get_origin(type_)
                if origin is None:
                    if issubclass(type_, cls.TYPES_ARGS):
                        continue
                    elif issubclass(type_, JsonApiData):
                        checker(get_type_hints(JsonApiData).values())
                    else:
                        raise InvalidAnnotationError(
                            f"{type_.__name__} is an unacceptable as a type "
                            "of attributes of JsonApiData.")
                else:
                    args = get_args(type_)
                    if origin == list:
                        if not len(args):
                            raise InvalidAnnotationError(
                                "List with no arguments is not acceptable. "
                                "Specify an argument of type to list.")
                        checker([args[0]])
                    elif origin == Union:
                        # Only Optional is admitted.
                        if not (len(args) == 2 and args[1] == cls.NoneType):
                            raise InvalidAnnotationError(
                                "Attritbutes with serveral types are not "
                                "acceptable. Only 'Optional' can be used as "
                                "type hints.")
                        
                        checker([args[0]])
                    else:
                        raise InvalidAnnotationError(
                            f"{type_.__name__} is not acceptable as a type of"
                            " attributes of JsonApiData.")
        checker(types)
        
    @classmethod
    def has_valid_annotations(cls, apiclass: Type[JsonApiData]) -> None:
        """Check if `JsonApiData` has only arguments convertable to JSON.

        Parameters
        ----------
        apiclass : Type[JsonApiData]
            A subclass of the `JsonApiData`
            
        Raises
        ------
        InvalidAnnotationError
            Raised if a type not convertable to JSON data is detected
        """
        cls.check_annotations(*tuple(get_type_hints(apiclass).values()))
    
    @classmethod
    def validate_obj(cls, obj: Any, objtype: Type) -> bool:
        """Validate if given object is acceptable as a value of `JsonApiData`.

        Parameters
        ----------
        obj : Any
            Value to be validated
        objtype : Type
            Type to be validated

        Returns
        -------
        bool
            If the obj and objtype is acceptable to `JsonApiData`
        """
        origin = get_origin(objtype)
        if origin is None:
            if objtype in cls.TYPES_ARGS_SET:
                return isinstance(obj, objtype)
            return issubclass(objtype, JsonApiData) and isinstance(obj, dict)
        
        if origin == list:
            if len(obj):
                item = obj[0]
                inner_type = get_args(objtype)[0]
                return cls.validate_obj(item, inner_type)
            return True
        elif origin == Union:
            is_good = False
            for arg in get_args(objtype):
                is_good = is_good or cls.validate_obj(obj, arg)
            return is_good
        else:
            return False
    
    @classmethod
    def build(cls, apiclass: Type[JsonApiData], 
              data: Dict[str, Any]) -> JsonApiData:
        """Build new `JsonApiData` object from raw data.

        Parameters
        ----------
        apiclass : Type[JsonApiData]
            Class of new object to be generated
        data : Dict[str, Any]
            Raw data converted as a dictionary

        Returns
        -------
        JsonApiData
            New object built

        Raises
        ------
        ValidationFailedError
            Raised if type validation of raw data failes
        """
        instance = apiclass.__new__(apiclass)
        annotations = get_type_hints(apiclass)
        
        # NOTE
        #   Ignore keys of data which is not defined in the apiclass.
        for key_def, type_def in annotations.items():            
            if not (key_def in data or hasattr(apiclass, key_def)):
                raise ValidationFailedError(
                    f"Raw data has no key of '{key_def}'.")
            
            val = data.get(key_def)
            if not cls.validate_obj(val, type_def):
                raise ValidationFailedError(
                    "Invalid type was detected in received json data. "
                    f"Expected: {type_def.__name__}; "
                    f"Detected: {val.__class__.__name__}")
            
            # Set values of attributes to an instance of the apiclass.
            origin = get_origin(type_def)
            if origin is None:       
                if issubclass(type_def, JsonApiData):
                    setattr(instance, key_def, cls.build(type_def, val))
                else:
                    setattr(instance, key_def, val)
            else:
                if origin == list:
                    setattr(instance, key_def, cls._build_list(type_def, val))
                elif origin == Union:
                    possible = get_args(type_def)[0]
                    if possible in cls.TYPES_ARGS_SET:
                        setattr(instance, key_def, val)
                    elif possible == list:
                        setattr(instance, key_def,
                                cls._build_list(possible, val))
                    elif issubclass(possible, JsonApiData):
                        setattr(instance, key_def, cls.build(type_def, val))
                    else:
                        raise ValidationFailedError(
                            f"{possible.__name__} is not acceptable as a type"
                            " of attributes of JsonApiData.")
        return instance
    
    @classmethod
    def _build_list(cls, type_def: ListJsonable_t, val: list
                    ) -> ListJsonable_t:
        """Build list convertable to JSON data.

        Parameters
        ----------
        type_def : ListJsonable_t
            Type of object to be generated, must be list with an argument
        val : list
            Raw data converted as a list

        Returns
        -------
        ListJsonable_t
            Generated object
            
        Raises
        ------
        ValueError
            Raised if `type_def` is not list with an argument
        ValidationFailedError
            Raised if the argument of `type_def` is not acceptable
        """
        if not len(val):
            return val
        if not(get_origin(type_def) == list and len(get_args(type_def))):
            raise ValueError("'type_def' must be list with an argument.")

        # Check type of inner value
        inner_type = get_args(type_def)[0] 
        if inner_type in cls.TYPES_ARGS_SET:
            return val
        elif issubclass(inner_type, JsonApiData):
            return [cls.build(inner_type, item) for item in val]
        elif get_origin(inner_type) == list:
            return [cls._build_list(inner_type, item) for item in val]
        else:
            raise ValidationFailedError(
                f"{inner_type.__name__} is not acceptable as a type of "
                "attributes of JsonApiData.")

    
class JsonApiData(ApiData):
    """API data with JSON format.

    This class can be used to describe data with JSON format. This class 
    should be inheritted and its several class-attributes should be 
    defined in the subclass. Developers must define type hints of the 
    class-attribtues, which are used to validate if raw data has format 
    the type hints define.
    
    Examples
    --------
    - Defining subclass of this class
    
    ```
    class User(JsonApiData):
        name: str
        email: str
        age: int
        
    class MockApiData(JsonApiData):
        users: List[User]
    ```
    
    - Validating received data
    
    ```
    class MockEndpoint(Endpoint):
    
        @data_format(input=MockApiData, output=None)
        def do_GET(self, rec_body: MockApiData) -> None:
            # Do something...
            
            # Example
            for user in rec_body.users:
                print(f"user name : {user.name}")
    ```
    """
    
    def __init_subclass__(cls) -> None:
        JsonApiDataBuilder.has_valid_annotations(cls)
            
    # NOTE
    #   DO NOT override the method. This class should be used only 
    #   by defining attributes and their types.
    def __init__(self, raw: bytes, content_type: ContentType) -> None:
        """
        Parameters
        ----------
        raw : bytes
            Raw data to be validated
        content_type : ContentType
            Values of `Content-Type` header
        """
        super().__init__(raw, content_type)
        
        encoding = content_type.charset
        if encoding is None:
            encoding = "UTF-8"
            
        media_type = content_type.media_type
        if media_type is None:
            raise ValidationFailedError(
                "'Content-Type' header was not specified. Specify "
                f"'{MediaTypes.json}' as media type.")
        if media_type.lower() != MediaTypes.json:
            raise ValidationFailedError(
                "Media type of 'Content-Type' header was not "
                f"{MediaTypes.json}, but {content_type.media_type}.")

        try:
            raw = raw.decode(encoding=encoding)
            data = json.loads(raw)
        except UnicodeDecodeError:
            raise ValidationFailedError(
                "Decoding raw data failed. The encoding was expected "
                f"{encoding}, but not corresponded.")
        except json.decoder.JSONDecodeError:
            raise ValidationFailedError(
                "Decoding raw data failed."
                "The raw data had invalid JSON format.")

        self.__dict__.update(
            JsonApiDataBuilder.build(self.__class__, data).__dict__)
        
    @class_property
    def _content_type_(cls) -> str:
        return MediaTypes.json


class XWWWFormUrlEncodedDataBuilder:
    
    TYPES_ARGS = (str,)
    
    @classmethod
    def check_annotations(cls, *types: Type) -> None:
        """Check if types are acceptable to `XWWWFormUriEncodedData` class.

        Parameters
        ----------
        *types : Type
            Classes to be checked

        Raises
        ------
        InvalidAnnotationError
            Raised if a type not acceptable
        """
        for type_ in types:
            if type_ not in cls.TYPES_ARGS:
                raise InvalidAnnotationError(
                    f"{type_.__name__} is an unacceptable as a type "
                    "of attributes of XWWWFormUrlEncodedData.")
            
    @classmethod
    def has_valid_annotations(cls, apiclass: Type[XWWWFormUrlEncodedData]
                              ) -> None:
        """Check if `XWWWFormUrlEncodedData` has only arguments with 
        acceptable types.

        Parameters
        ----------
        apiclass : Type[XWWWFormUrlEncodedData]
            A subclass of the `XWWWFormUrlEncodedData`
            
        Raises
        ------
        InvalidAnnotationError
            Raised if a type not acceptable
        """
        cls.check_annotations(*tuple(get_type_hints(apiclass).values()))
    
    @classmethod
    def build(cls, apiclass: Type[XWWWFormUrlEncodedData],
              data: str) -> XWWWFormUrlEncodedData:
        """Build new `XWWWFormUrlEncodedData` object from raw data.

        Parameters
        ----------
        apiclass : Type[XWWWFormUrlEncodedData]
            Class of new object to be generated
        data : str
            Raw data already decoded

        Returns
        -------
        XWWWFormUrlEncodedData
            New object built

        Raises
        ------
        ValidationFailedError
            Raised if type validation of raw data failes
        """
        instance = apiclass.__new__(apiclass)
        annotations = get_type_hints(apiclass)        
        data = parse_qs(data)
        
        for key_def in annotations.keys():
            if key_def not in data:
                raise ValidationFailedError(
                    f"Raw data has no key of '{key_def}'.")
                
            if len(data[key_def]) > 1:
                raise ValidationFailedError(
                    f"Duplicated keys '{key_def}' were detected.")
            
            setattr(instance, key_def, data[key_def][0])
        
        return instance
        

class XWWWFormUrlEncodedData(ApiData):
    """API data with `x-www-form-urlencoded`
    
    This class can be used to describe data with `x-www-form-urlencoded` 
    format. This class should be inheritted and its several 
    class-attributes should be defined in the subclass. Developer must 
    define type hints of the class-attributes, which are used to validate if 
    raw data has format the type hints define. In this class, the type hints 
    must be only `str`. Otherwise, `InvalidAnnotationError` will be raised.

    Examples
    --------
    - Defining subclass of this class
    
    ```
    class UserCredentials(XWWWFormUrlEncodedData):
        user_id: str
        password: str
    ```
    
    - Validating received data
    
    ```
    class MockEndpoint(Endpoint):
    
        @data_format(input=UserCredentials, output=None)
        def do_GET(self, rec_body: UserCredentials) -> None:
            # Do something...
            
            # Example
            authenticate(rec_body.user_id, rec_body.password)
    ```
    """
    
    def __init_subclass__(cls) -> None:
        XWWWFormUrlEncodedDataBuilder.has_valid_annotations(cls)
    
    def __init__(self, raw: bytes, content_type: ContentType) -> None:
        """
        Parameters
        ----------
        raw : bytes
            Raw data to be validated
        content_type : ContentType
            Values of `Content-Type` header
        """
        encoding = content_type.charset
        if encoding is None:
            encoding = "UTF-8"
            
        media_type = content_type.media_type
        if media_type is None:
            raise ValidationFailedError(
                "'Content-Type' header was not specified. Specify "
                f"'{MediaTypes.x_www_form_urlencoded}' as media type.")
        if media_type.lower() != MediaTypes.x_www_form_urlencoded:
            raise ValidationFailedError(
                "Media type of 'Content-Type' header is not "
                f"{MediaTypes.x_www_form_urlencoded}, "
                f"but {content_type.media_type}.")
        
        try:
            raw = raw.decode(encoding=encoding)
        except UnicodeDecodeError:
            raise ValidationFailedError(
                "Decoding raw data failed. The encoding was expected "
                f"{encoding}, but not corresponded.")
        
        instance = XWWWFormUrlEncodedDataBuilder.build(self.__class__, raw)
        self.__dict__.update(instance.__dict__)
    
    @class_property
    def _content_type_(cls) -> str:
        return MediaTypes.x_www_form_urlencoded




from __future__ import annotations
from http.client import HTTPConnection, HTTPMessage, HTTPResponse
from typing import Generic, Optional, Type

from bamboo.api import BinaryApiData
from bamboo.base import ContentType
from bamboo.request import ResponseData_t
from bamboo.util.deco import cached_property


class Response(Generic[ResponseData_t]):
    
    def __init__(self,
                 conn: HTTPConnection,
                 res: HTTPResponse, 
                 datacls: Type[ResponseData_t] = BinaryApiData
                 ) -> None:
        self._conn = conn
        self._res = res
        self._datacls = datacls

    @property
    def headers(self) -> HTTPMessage:
        return self._res.msg
    
    def get_header(self, key: str) -> Optional[str]:
        return self._res.getheader(key)
    
    @property
    def url(self) -> str:
        return self._res.geturl()
    
    @property
    def status(self) -> int:
        return self._res.status
    
    @property
    def version(self) -> int:
        return self._res.version
    
    @property
    def ok(self) -> bool:
        status = self.status
        if 200 <= status < 300:
            return True
        return False
    
    @property
    def is_closed(self) -> bool:
        return self._res.closed
    
    @property
    def fileno(self) -> int:
        return self._res.fileno()
    
    @property
    def content_length(self) -> Optional[int]:
        length = self.get_header("Content-Length")
        if length is not None:
            return int(length)
        return None
    
    @cached_property
    def body(self) -> bytes:
        return self._res.read(self.content_length)
    
    def attach(self, 
               datacls: Optional[Type[ResponseData_t]] = None
               ) -> ResponseData_t:
        content_type_raw = self.get_header("Content-Type")
        if content_type_raw:
            content_type = ContentType.parse(content_type_raw)
        else:
            content_type = ContentType()

        if datacls is None:
            return self._datacls(self.body, content_type)
        else:
            return datacls(self.body, content_type)
    
    def close(self) -> None:
        self._conn.close()
        
    def __enter__(self) -> Response:
        return self
    
    def __exit__(self, type, value, traceback) -> None:
        self.close()
    